# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Exceptions used by WereSync"""


class DeviceError(Exception):
    """Exception thrown to show errors caused by an issue with a specific
    device."""

    def __init__(self, device, message, errors=None):
        self.device = device
        self.message = message
        self.errors = errors


class CopyError(Exception):
    """Exception thrown to show errors caused by an issue copying data,
    usually both devices face the issue."""

    def __init__(self, message, errors=None):
        self.message = message
        self.errors = errors


class UnsupportedDeviceError(Exception):
    """Exception thrown to show that action is not supported on the partition
    table type of the device."""
    pass


class InvalidVersionError(Exception):
    """Exception thrown when the version of python being used does not support
    the feature."""
    pass


class PluginNotFoundError(Exception):
    """Exception thrown when the passed plugin is not found."""
    pass



#! /usr/bin/env python3
# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Runs the daemon and exposes the necessary methods to the interface
functions.

See also :mod:`weresync.interface.dbus_client`:"""

from weresync.daemon.copier import DriveCopier
from gi.repository import GLib
from pydbus import SystemBus
import weresync.utils as utils
import logging

DEFAULT_DAEMON_LOG_LOCATION = "/var/log/weresync/weresync.log"

LOGGER = logging.getLogger(__name__)


def run():
    """Function which starts the daemon and publishes the
    :class:`~weresync.daemon.copier.DriveCopier` over dbus."""
    utils.start_logging_handler(DEFAULT_DAEMON_LOG_LOCATION)
    utils.enable_localization()
    with SystemBus() as bus:
        with bus.publish("net.manilas.weresync.DriveCopier", DriveCopier()):
            GLib.idle_add(lambda: LOGGER.debug("Starting GLib loop"))
            loop = GLib.MainLoop()
            loop.run()


if __name__ == "__main__":
    run()



import weresync.daemon.device as device
import weresync.utils as utils
from weresync.exception import (CopyError, DeviceError, UnsupportedDeviceError)
import random
import os
import logging
import subprocess
from pydbus.generic import signal

LOGGER = logging.getLogger(__name__)


def mount_loop_device(image_file):
    """Mounts an image file as a loop device and returns the device name of
    the mounted loop. This mounts on first free loop device. This accepts
    relative paths.

    :params image_file: Path pointing to the image file to mount.
    :returns: A string containing device identifier (/dev/sda or such)"""

    image_file = os.path.abspath(os.path.expanduser(image_file))
    free_proc = subprocess.Popen(
        ["losetup", "-f"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    free_output, free_error = free_proc.communicate()
    if free_proc.returncode != 0:
        raise DeviceError(image_file, "Error finding free loop device.",
                          str(free_output, "utf-8"))

    device_name = str(free_output, "utf-8").strip()
    mount_proc = subprocess.Popen(
        ["losetup", device_name, image_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)
    mount_output, mount_error = mount_proc.communicate()
    if mount_proc.returncode != 0:
        raise DeviceError(image_file,
                          "Error mounting image on {0}".format(device_name),
                          str(mount_output, "utf-8"))
    subprocess.call(["partprobe", device_name])
    return device_name


def create_new_vg_if_not_exists(lvm, name, target):
    """Creates a new Logical Volume Group with the name ``lvm`` + "copy"
    and all of the partitions of the target with type "lvm" added to it.

    This is not a conclusive function and misses several uses of LVM drives,
    if your situation is not covered, please feel free to open a pull request
    fixing it.

    :param lvm: a string representing the name of the source lvm
    :param target: a :py:class:`~weresync.device.DeviceManager` representing
                   the device whose partitions to add to the LVM."""
    lvm_partitions = []
    for i in target.get_partitions():
        code = target.get_partition_code(i).lower()
        if code == "8e00" or code == "8e":
            # the two versions appear in gdisk and fdisk, respectively
            lvm_partitions.append(i)

    lvm_part_block = [
        target.part_mask.format(target.device, x) for x in lvm_partitions
    ]
    lvm_test = subprocess.Popen(
        ["vgs", name], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output, _ = lvm_test.communicate()
    if lvm_test.returncode == 5:
        # If the VG does not exist, the return code is 5
        utils.run_proc(["vgcreate", name] + lvm_part_block, target.device,
                       "Error creating logical volume group.")
    else:
        if name.startswith("/dev/"):
            name = name[5:]
        result = utils.run_proc(
            ["pvs", "-S", "vg_name=" + name, "--noheadings", "-o", "pv_name"],
            name, "Error finding physical volumes for LVM.")
        LOGGER.debug("PVs in LVM: " + result)
        results = [x.strip() for x in result.split("\n")]
        lvm_part_block = [x for x in lvm_part_block if x not in results]
        if len(lvm_part_block) > 0:
            utils.run_proc(["vgextend", name] + lvm_part_block, name,
                           "Error adding PVs to LVM")


def copy_partitions(copier, part_callback=None, lvm=False):
    """Checks if partitions are valid and copies if they aren't.

    :param copier: the :py:class:`~weresync.device.DeviceCopier` object to do
                   the copying with.
    :param part_callback: see the documentation for :py:func:`~.copy_drive`"""
    try:
        print(_("Checking partition validity."))
        copier.partitions_valid(lvm)
        if part_callback is not None:
            part_callback(1.0)
            LOGGER.info("Drives are compatible")
    except CopyError as ex:
        LOGGER.warning(ex.message)
        print(_("Partitions invalid!\nCopying drive partition table."))
        LOGGER.warning("Drives are incompatible.")
        if lvm:
            copier.transfer_lvm_partition(callback=part_callback)
        else:
            copier.transfer_partition_table(callback=part_callback)
    else:
        if part_callback is not None:
            part_callback(1.0)


class DriveCopier(object):
    """Object which is shared over dbus to allow interfaces to access the
    daemon."""

    def __init__(self):
        LOGGER.debug("making object")
        self._interface_name = "net.manilas.weresync.DriveCopier"

    dbus = """<node>
      <interface name='net.manilas.weresync.DriveCopier'>
        <method name='CopyDrive'>
          <arg type='s' name='source' direction='in' />
          <arg type='s' name='target' direction='in' />
          <arg type='b' name='check_if_valid_and_copy' direction='in' />
          <arg type='s' name='source_part_mask' direction='in' />
          <arg type='s' name='target_part_mask' direction='in' />
          <arg type='ai' name='excluded_partitions' direction='in' />
          <arg type='b' name='ignore_copy_failures' direction='in' />
          <arg type='i' name='root_partition' direction='in' />
          <arg type='i' name='boot_partition' direction='in' />
          <arg type='i' name='efi_partition' direction='in' />
          <arg type='(ss)' name='mount_points' direction='in' />
          <arg type='s' name='rsync_args' direction='in' />
          <arg type='s' name='lvm_source' direction='in' />
          <arg type='s' name='lvm_target' direction='in' />
          <arg type='s' name='bootloader' direction='in' />
          <arg type='s' name='return' direction='out' />
        </method>
        <method name='GetPartitions'>
          <arg type='s' name='device' direction='in' />
          <arg type='s' name='part_mask' direction='in' />
          <arg type='b' name='lvm' direction='in' />
          <arg type='as' name='partitions' direction='out' />
        </method>
        <signal name="PartitionStatus">
          <arg type='d' name='percent_complete' direction='out' />
        </signal>
        <signal name="CopyStatus">
          <arg type='d' name='percent_complete' direction='out' />
          <arg type='i' name='current_partition' direction='out' />
        </signal>
        <signal name="BootStatus">
          <arg type='b' name='completed' direction='out' />
        </signal>
      </interface>
    </node>"""

    PartitionStatus = signal()
    CopyStatus = signal()
    BootStatus = signal()

    def GetPartitions(self, device_name, part_mask, lvm):
        """Gets the partitions of the specified device. Note that this
        will always return a list of strings, however if this is not an LVM
        device, all of the strings will be valid numbers."""
        if lvm:
            source_manager = device.LVMDeviceManager(device_name)
        else:
            source_manager = device.DeviceManager(device_name, part_mask)
        parts = source_manager.get_partitions()
        for idx, val in enumerate(parts):
            if type(val) != str:
                parts[idx] = str(val)
        return parts

    def CopyDrive(self,
                  source,
                  target,
                  check_if_valid_and_copy=False,
                  source_part_mask="{0}{1}",
                  target_part_mask="{0}{1}",
                  excluded_partitions=[],
                  ignore_copy_failures=True,
                  root_partition=-1,
                  boot_partition=-1,
                  efi_partition=-1,
                  mount_points=("", ""),
                  rsync_args=device.DEFAULT_RSYNC_ARGS,
                  lvm_source="",
                  lvm_target="",
                  bootloader="uuid_copy"):
        """Uses a DeviceCopier to clone the source drive to the target drive.

        **Note:** if using LVM, any uses of "partition" in the documentation
        actually refer to logical volumes.

        It is recommended to set ``check_if_valid_and_copy`` to True if the the
        two drives are not the same size with the same partitions.

        If either source or target ends in ".img" copy_drives will assume it is an
        image file, and mount if accordingly.

        :param source: The drive identifier ("/dev/sda" or the like) of the source
                       drive.
        :param target: The drive identifier ("/dev/sda" or the like) of the target
                       drive.
        :param check_if_valid=False: If true, the function checks if the target
                                     drive is compatible to receive the source
                                     drive's data. If it is not, erase the target
                                     drive and make a proper partition table.
                                     Defaults to False.
        :param source_part_mask: A string to be passed to the "format" method that
                                 expects to arguments, the drive name and the
                                 partition number. Applied to the source drive.
                                 Defaults to "{0}{1}".
        :param target_part_mask: Same as source_part_mask, but applied to target
                                 drive. Defaults to "{0}{1}"
        :param excluded_partitions: Partitions to not copy or test for boot
                                    capability.
        :param ignore_copy_failures: If True, errors during copying will be
                                     ignored and copying will continue. It is
                                     recommended that this be left to true,
                                     because errors frequently occur with swap
                                     partitions or other strange partitions.
        :param root_partition: If not None, this is an int that determines which
                               partition grub should be installed to. Defaults to
                               None.
        :param boot_partition: If not None, this is an int that represents the
                               partition to mount at /boot when installing grub.
        :param efi_partition: If not None, this is an int that represents the
                              partition to mount at /boot/efi when installing grub.
        :param mount_points: Expects a tuple containing two strings pointing to
                             the directories where partitions should be mounted in
                             case of testing. If None, the function will generate
                             two random directories in the /tmp folder. Defaults
                             to None.
        :param lvm: the Logical Volume Group to copy to the new drive.

        :raises DeviceError: If there is an error reading data from one device or
                             another.
        :raises CopyError: If there is an error copying the data between the two
                           devices.

        :returns: True on success and an error message or exception on failure.
        """
        LOGGER.debug("Daemon starting clone.")

        def part_callback(status):
            self.PartitionStatus(status)

        def copy_callback(current, status):
            self.CopyStatus(current, status)

        def boot_callback(status):
            self.BootStatus(status)

        root_partition = root_partition if root_partition >= 0 else None
        boot_partition = boot_partition if boot_partition >= 0 else None
        efi_partition = efi_partition if efi_partition >= 0 else None

        try:
            source_loop = None
            target_loop = None
            if source.endswith(".img"):
                source_loop = mount_loop_device(source)
                source = source_loop
                source_part_mask = "{0}p{1}"

            if target.endswith(".img"):
                target_loop = mount_loop_device(target)
                target = target_loop
                target_part_mask = "{0}p{1}"
                LOGGER.warning(
                    "Right now, WereSync does not properly install bootloaders on "
                    "image files. You will have to handle that yourself if you "
                    "want your image to be bootable.")

            source_manager = device.DeviceManager(source, source_part_mask)
            target_manager = device.DeviceManager(target, target_part_mask)

            try:
                target_manager.get_partition_table_type()
            except (DeviceError, UnsupportedDeviceError) as ex:
                # Since we're erasing the target drive anyway, we can just create
                # a new disk label
                proc = subprocess.Popen(
                    ["sgdisk", "-o", target_manager.device])
                proc.communicate()

            copier = device.DeviceCopier(source_manager, target_manager)
            partitions_remade = False
            if check_if_valid_and_copy:
                copy_partitions(copier, part_callback)
                partitions_remade = True

            if lvm_source is not "":
                create_new_vg_if_not_exists(lvm_source, lvm_source + "-copy",
                                            target_manager)
                lvm_source = device.LVMDeviceManager(lvm_source)
                lvm_target = device.LVMDeviceManager(lvm_source.device +
                                                     "-copy")
                copier.lvm_source = lvm_source
                copier.lvm_target = lvm_target
                if partitions_remade and check_if_valid_and_copy:
                    copy_partitions(copier, part_callback, lvm=True)

            if mount_points is ("", "") or len(
                    mount_points) < 2 or mount_points[0] == mount_points[1]:
                source_dir = "/tmp/" + str(random.randint(0, 100000))
                target_dir = "/tmp/" + str(random.randint(-100000, -1))
                os.makedirs(source_dir, exist_ok=True)
                os.makedirs(target_dir, exist_ok=True)
                mount_points = (source_dir, target_dir)

            print(_("Beginning to copy files."))
            copier.copy_files(
                mount_points[0],
                mount_points[1],
                excluded_partitions,
                ignore_copy_failures,
                rsync_args,
                callback=copy_callback)
            print(_("Finished copying files."))

            print(_("Making bootable"))
            try:
                copier.make_bootable(bootloader, mount_points[0],
                                     mount_points[1], excluded_partitions,
                                     root_partition, boot_partition,
                                     efi_partition, boot_callback)
            except DeviceError as ex:
                print(
                    _("Error making drive bootable. All files should be fine.")
                )
                return ex
            print(_("All done, enjoy your drive!"))
            return "True"
        finally:

            def delete_loop(loop_name):
                subprocess.call(["losetup", "-d", loop_name])

            if source_loop is not None:
                delete_loop(source_loop)
            if target_loop is not None:
                delete_loop(target_loop)



#! /usr/bin/env python3
# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This modules has easy, one function interfaces with the DeviceCopier and
DeviceManager."""

import weresync.daemon.device as device
from weresync.exception import InvalidVersionError
import weresync.interface.dbus_client as dbus_client
import logging
import logging.handlers
import sys
import argparse
import weresync.utils as utils

LOGGER = logging.getLogger(__name__)

DEFAULT_LENGTH = 60

partition_copying_completed = False


def _print_progress_bar(iteration,
                        total,
                        prefix='',
                        suffix='',
                        decimals=1,
                        length=DEFAULT_LENGTH,
                        fill='█'):
    """
    Call in a loop to create terminal progress bar

    Thanks to Greentick from StackOverflow for this function
    https://stackoverflow.com/a/34325723/2719960
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
    """
    # the first part of the above tuple is the terminal width
    percent = ("{0:." + str(decimals) + "f}").format(
        100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end='\r')
    # Print New Line on Complete
    if iteration == total:
        print()


def main():
    """The entry point for the command line function. This uses argparse to
    parse arguments to call call :py:func:`.copy_drive` with. For help use
    "weresync -h" in a commandline after installation."""

    if not dbus_client.copy_drive:
        LOGGER.error("Failed to connect to dbus service. Is it running?")
        LOGGER.log(logging.DEBUG, "DBus Connection error:",
                   dbus_client.error)
        return

    def part_callback(status):
        global partition_copying_completed
        if not partition_copying_completed:
            _print_progress_bar(
                int(status * 100), 100, prefix=_("Partition Table Copying:"))
        if status >= 1:
            partition_copying_completed = True

    def copy_callback(partition_num, status):
        _print_progress_bar(
            int(status * 100) if status >= 0 else 0,
            100,
            prefix=_("Copying Partition {0}:").format(partition_num))
        if status < 0:
            print()
            print(
                _("Some error copying partition {0}. "
                  "Probably fine, but see logs.").format(partition_num))

    def boot_callback(status):
        if status:
            print(_("Bootloader installed"))
        # Passes true when completed. CMD doesn't have display for
        # incomplete installation so only test if True

    utils.enable_localization()
    dbus_client.subscribe_to_signals(part_callback, copy_callback,
                                     boot_callback)

    try:
        utils.check_python_version()
    except InvalidVersionError as ex:
        print(ex)
        sys.exit(1)

    try:
        import weresync.plugins as plugins
        manager = plugins.get_manager()
        manager.collectPlugins()
        default_part_mask = "{0}{1}"
        pluginNames = []
        for pluginInfo in manager.getAllPlugins():
            pluginNames.append(pluginInfo.plugin_object.name)
        epilog_string = (
            _("Bootloader plugins found: ") + ", ".join(pluginNames))
        parser = argparse.ArgumentParser(epilog=epilog_string)
        parser.add_argument(
            "source",
            help=_("The drive to copy data from. This drive will not be"
                   " edited."))
        parser.add_argument(
            "target",
            help=_("The drive to copy data to. ALL DATA ON THIS DRIVE WILL BE "
                   "ERASED."))
        parser.add_argument(
            "-C",
            "--check-and-partition",
            action="store_true",
            help=_("Check if partitions are valid and re-partition drive to "
                   "proper partitions if they are not."))
        parser.add_argument(
            "-s",
            "--source-mask",
            help=_("A string of format '{0}{1}' where {0} represents drive "
                   "identifier and {1} represents partition number to point "
                   "to partition block files for the source drive."))
        parser.add_argument(
            "-t",
            "--target-mask",
            help=_("A string of format '{0}{1}' where {0} represents drive "
                   "identifier and {1} represents partition number to point "
                   "to partition block files for the target drive."))
        parser.add_argument(
            "-e",
            "--excluded-partitions",
            help=_("A comment separated list of partitions of the source "
                   "drive to apply no actions on.perated list of partitions "
                   "of the source drive to apply no actions on."))
        parser.add_argument(
            "-b",
            "--break-on-error",
            action="store_false",
            help=_("Causes program to break whenever a partition cannot be "
                   "copied, including uncopyable partitions such as swap"
                   " files. Not recommended."))
        parser.add_argument(
            "-g",
            "--root-partition",
            type=int,
            help=_("The partition mounted on /."),
            default=-1)
        parser.add_argument(
            "-B",
            "--boot-partition",
            type=int,
            help=_("Partition which should be mounted on /boot"),
            default=-1)
        parser.add_argument(
            "-E",
            "--efi-partition",
            type=int,
            help=_("Partition which should be mounted on /boot/efi"),
            default=-1)
        parser.add_argument(
            "-m",
            "--source-mount",
            help=_("Folder where partitions from the source drive should be "
                   "mounted."),
            default="")
        parser.add_argument(
            "-M",
            "--target-mount",
            help=_("Folder where partitions from source drive should be"
                   " mounted."),
            default="")
        parser.add_argument(
            "-r",
            "--rsync-args",
            help=_("List of arguments passed to rsync. Defaults to: ") +
            device.DEFAULT_RSYNC_ARGS,
            default=device.DEFAULT_RSYNC_ARGS)
        parser.add_argument(
            "-L",
            "--bootloader",
            help=_("Passed to decide what boootloader plugin to use. See "
                   "below for list of plugins. Defaults to simply changing "
                   "the UUIDs of files in /boot."),
            default="uuid_copy")
        parser.add_argument(
            "-l",
            "--lvm",
            help=_("The name of the source logical volume."),
            nargs="+")
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-v",
            "--verbose",
            help=_("Prints expanded output."),
            action="store_const",
            dest="loglevel",
            const=logging.INFO)
        group.add_argument(
            "-d",
            "--debug",
            help=_("Prints large output. Mainly helpful for developers."),
            action="store_const",
            dest="loglevel",
            const=logging.DEBUG)
        args = parser.parse_args()
        if (args.loglevel == logging.INFO or args.loglevel == logging.DEBUG):
            loglevel = args.loglevel
        else:
            loglevel = logging.WARNING
        utils.start_logging_handler(
            utils.DEFAULT_USER_LOG_LOCATION, stream_level=loglevel)
        mount_points = (args.source_mount, args.target_mount)
        if args.lvm is not None and len(args.lvm) > 2:
            LOGGER.warning(
                _("More than two lvm options added. Please give "
                  "either one or two options."))
            sys.exit(1)

        lvm_source = args.lvm[0] if args.lvm is not None else ""
        lvm_target = args.lvm[1] if (args.lvm is not None
                                     and len(args.lvm) == 2) else ""

        excluded_partitions = [
            int(x) for x in args.excluded_partitions.split(",")
        ] if args.excluded_partitions is not None else []
        source_mask = (args.source_mask
                       if args.source_mask is not None else default_part_mask)
        target_mask = (args.target_mask
                       if args.target_mask else default_part_mask)
        result = dbus_client.copy_drive(
            args.source, args.target, args.check_and_partition,
            source_mask, target_mask, excluded_partitions,
            args.break_on_error, args.root_partition, args.boot_partition,
            args.efi_partition, mount_points, args.rsync_args, lvm_source,
            lvm_target, args.bootloader)
        if result == "True":
            print(_("All done, enjoy your drive!"))
        else:
            print(str(result))

    except (KeyboardInterrupt, EOFError):
        LOGGER.info("Exiting via user keyboard interrupt.")
        LOGGER.debug("Error:\n", exc_info=sys.exc_info())
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Installs the Grub2 bootloader. This works on both UEFI and MBR systems."""


from weresync.plugins import IBootPlugin
import weresync.plugins as plugins
import weresync.daemon.device as device
from weresync.exception import CopyError, DeviceError
import subprocess
import os
import sys
import logging

LOGGER = logging.getLogger(__name__)


class GrubPlugin(IBootPlugin):
    """Plugin to install the grub2 bootloader. Does not install grub legacy."""

    def __init__(self):
        super().__init__("grub2", "Grub2")

    def get_help(self):
        return __doc__

    def install_bootloader(self, source_mnt, target_mnt, copier,
                           excluded_partitions=[],
                           boot_partition=None, root_partition=None,
                           efi_partition=None):

        if efi_partition is not None:
            import weresync.plugins.weresync_uuid_copy as uuc
            # UEFI systems tend to only need a UUID copy. No sense in not
            # reusing old code.
            uuc.UUIDPlugin().install_bootloader(source_mnt, target_mnt, copier,
                                                excluded_partitions,
                                                boot_partition,
                                                root_partition, efi_partition)
            return

        if root_partition is None and boot_partition is None:
            # This for loop searches for a partition with a /boot/grub folder
            # and it assumes it is the root partition
                for i in copier.target.get_partitions():
                    try:
                        mount_point = copier.target.mount_point(i)
                        if mount_point is None:
                            copier.target.mount_partition(i, target_mnt)
                            mount_point = target_mnt
                        if os.path.exists(mount_point +
                                          ("/" if not mount_point.endswith("/")
                                           else "") + "boot/grub"):
                            root_partition = i
                            break
                        else:
                            copier.target.unmount_partition(i)
                    except DeviceError as ex:
                        LOGGER.warning("Could not mount partition {0}. "
                                       "Assumed to not be the partition grub "
                                       "is on.".format(i))
                        LOGGER.debug("Error info:\n", exc_info=sys.exc_info())
                else:  # No partition found
                    raise CopyError("Could not find partition with "
                                    "'boot/grub' folder on device {0}".format(
                                        copier.target.device))

        # These variables are flags that allow the plugin to know if it mounted
        # any partitions and then clean up properly if it did
        mounted_here = False
        boot_mounted_here = False
        try:
            if root_partition is not None:
                mount_loc = copier.target.mount_point(root_partition)
                if mount_loc is None:
                    plugins.mount_partition(copier.target, copier.lvm_target,
                                            root_partition, target_mnt)
                    mounted_here = True
                    mount_loc = target_mnt
            else:
                mount_loc = target_mnt

            # This line avoids double slashes in path
            mount_loc += "/" if not mount_loc.endswith("/") else ""

            if boot_partition is not None:
                boot_folder = mount_loc + "boot"
                if not os.path.exists(boot_folder):
                    os.makedirs(boot_folder)
                plugins.mount_partition(copier.target, copier.lvm_target,
                                        boot_partition, boot_folder)
                boot_mounted_here = True

            print(_("Updating Grub"))
            grub_cfg = mount_loc + "boot/grub/grub.cfg"
            old_perms = os.stat(grub_cfg)[0]
            try:
                with open(grub_cfg, "r+") as grubcfg:
                    cfg = grubcfg.read()
                    LOGGER.debug("UUID Dicts: " + str(copier.get_uuid_dict()))
                    final = device.multireplace(cfg, copier.get_uuid_dict())
                    grubcfg.seek(0)
                    grubcfg.write(final)
                    grubcfg.truncate()
                    grubcfg.flush()
            finally:
                os.chmod(grub_cfg, old_perms)

            print(_("Installing Grub"))
            grub_command = ["grub-install",
                            "--boot-directory=" + mount_loc + "boot",
                            "--recheck",
                            "--target=i386-pc", copier.target.device]
            LOGGER.debug("Grub command: " + " ".join(grub_command))

            grub_install = subprocess.Popen(grub_command,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.STDOUT)
            install_output, install_error = grub_install.communicate()
            if grub_install.returncode != 0:
                raise DeviceError(copier.target.device,
                                  "Error installing grub.",
                                  str(install_output,
                                      "utf-8"))

            print(_("Consider running update-grub on your backup. WereSync"
                  " copies can sometimes fail to capture all the nuances of a"
                  " complex system."))
            print(_("Cleaning up."))
        finally:
            # This block cleans up any mounted partitions
            if boot_mounted_here:
                copier.target.unmount_partition(boot_partition)
            if mounted_here:
                copier.target.unmount_partition(root_partition)
        print(_("Finished!"))



# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This modules contains the code to simply translate the UUIDs of all text
files to the new drive. It does not change anything else.

In order to save RAM, uuid_copy will not copy files larger than 200 MB.
This works for many bootloaders."""

from weresync.plugins import IBootPlugin
from weresync.exception import CopyError
import logging
import weresync.plugins as plugins

LOGGER = logging.getLogger(__name__)


class UUIDPlugin(IBootPlugin):

    def __init__(self):
        super().__init__("uuid_copy", "UUID Copy")

    def get_help(self):
        return """Changes all UUIDs in every file of /boot to the new drive's UUIDs.
        \nDoes not install anything else. This is the default option."""

    def install_bootloader(self, source_mnt, target_mnt, copier,
                           excluded_partitions=[],
                           boot_partition=None, root_partition=None,
                           efi_partition=None):

        if root_partition is None and boot_partition is None:
                part = plugins.search_for_boot_part(target_mnt, copier.target,
                                                    "boot",
                                                    excluded_partitions)
                if part is None:
                    raise CopyError("Could not find partition with "
                                    "'boot' folder on device {0}".format(
                                        copier.target.device))
                plugins.translate_uuid(copier, part, "/boot", target_mnt)
        elif boot_partition is not None:
            plugins.translate_uuid(copier, boot_partition, "/", target_mnt)
        else:
            plugins.translate_uuid(copier, root_partition, "/boot", target_mnt)

        if efi_partition is not None:
            plugins.translate_uuid(copier, efi_partition, "/", target_mnt)



# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains the classes used to modify block devices and copy
data."""

import shlex
import os
import subprocess
import weresync.exception
from weresync.exception import DeviceError, PluginNotFoundError
import math
import logging
import sys
import glob
import parse
import re
import tempfile

MOUNT_POINT = "/mnt"

LOGGER = logging.getLogger(__name__)
SUPPORTED_FILESYSTEM_TYPES = ["swap", "ef"]
for word in glob.glob("/*/mkfs.*"):
    val = word.split(".")
    SUPPORTED_FILESYSTEM_TYPES += val

SUPPORTED_PARTITION_TABLE_TYPES = ["gpt", "msdos"]
"""The names, as reported by `parted` of the partition table types supported
by this program."""
DEFAULT_RSYNC_ARGS = "-aAXxH --delete"
"""Default arguments passed to rsync. See rsync documentation for what they
do."""


def multireplace(string, replacements):
    """
    Given a string and a replacement map, it returns the replaced string.

    `Credit goes to bgusach
    <https://gist.github.com/bgusach/a967e0587d6e01e889fd1d776c5f3729>`_

    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary
                              {value to find: value to replace}
    :returns: a string with the replaced text."""
    # Place longer ones first to keep shorter substrings from matching where
    # the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'} against
    # the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    substrs = sorted(replacements, key=len, reverse=True)

    # Create a big OR regex that matches any of the substrings to replace
    regexp = re.compile('|'.join(map(re.escape, substrs)))

    # For each match, look up the new string in the replacements
    return regexp.sub(lambda match: replacements[match.group(0)], string)


class DeviceManager:
    """A class that allows various operations on a device.

    Most methods in this class raise
    :py:class:`~weresync.exception.DeviceError`
    if the subprocess they initiate has an error.

    :param device: a string containing the device identifier (/dev/sda or the
                   like)
    :param partition_mask: a string as the first parameter of a the expression:
                           "partion_mask.format(device, partition)" resolving
                           to a string. Defaults to "{0}{1}"."""

    def __init__(self, device, partition_mask="{0}{1}"):
        """Defines a device manager for the passed device.

        :param device: a string containing the device identifier (/dev/sda or
                       the like)
        :param partition_mask: a string as the first parameter of a the
                               expression: "partion_mask.format(device,
                               partition)" resolving to a string. Defaults to
                               "{0}{1}"."""
        self.device = device
        self.part_mask = partition_mask

    def get_partitions(self):
        """Returns a list with all the partitions in the drive. The partitions
        will be listed in **the order they appear on the disk**. So if
        partition 4 has a start sector of 500, it will appear before partition
        1 that has a start sectro of 2000

        :returns: A list of integers representing the partition numbers"""
        partition_table_proc = subprocess.Popen(
            ["parted", "-s", self.device, "print"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = partition_table_proc.communicate()
        exit_code = partition_table_proc.returncode
        if exit_code != 0:
            raise weresync.exception.DeviceError(self.device,
                                                 "Non-zero exit code",
                                                 str(output, "utf-8"))
        partition_result = str(output, "utf-8").split("\n")
        partitions = []
        for i in partition_result:
            line = i.strip().split()
            if len(line) == 0:
                continue  # It was an empty line
            num = None
            try:
                num = int(line[0])
            except ValueError:
                continue
            partitions.append(num)
        return partitions

    def mount_point(self, partition_num):
        """Returnds an absolute path to the mountpoint of the specific partition.
        Returns None if no mountpoint found (probably because partion not
        mounted). The return will not have a trailing slash.

        :param partition_num: The partition of whose mount to find."""

        findprocess = subprocess.Popen(
            [
                "findmnt", "-o", "TARGET", self.part_mask.format(self.device,
                                                                 partition_num)
            ],
            stdout=subprocess.PIPE)
        output, error = findprocess.communicate()
        result = str(output, "utf-8").split("\n")
        exit_code = findprocess.returncode
        if exit_code != 0 and exit_code != 1:
            # if nothing is found, findmnt returns 1; this is valid code
            raise weresync.exception.DeviceError(self.device,
                                                 "Non-zero exit code",
                                                 str(output, "utf-8"))

        if len(result) >= 2:
            return result[1].strip().split()[0]
        else:
            return None

    def mount_partition(self, partition_num, mount_loc):
        """Mounts the specified partition at the specified location.

        :param partition_num: the number of the partition to mount
        :param mount_loc: the location at which to mount the drive
        :raises: :py:class:`~weresync.exception.DeviceError` if there is an
                 error mounting the device.
        """
        mount_proc = subprocess.Popen(
            [
                "mount", self.part_mask.format(self.device, partition_num),
                mount_loc
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = mount_proc.communicate()
        exit_code = mount_proc.returncode
        if exit_code != 0:
            raise weresync.exception.DeviceError(
                self.device,
                "Non-zero exit code. Partition Number: {0}".format(
                    partition_num), str(output, "utf-8"))

        # if no error, mount succeeded

    def unmount_partition(self, partition_num):
        """Unmounts a device.

        :param partition_num: the number of the partition to unmount.
        :raises: :py:class:`~weresync.exception.DeviceError` if the partition
                 is busy or the partition is not mounted.
        """
        unmount_proc = subprocess.Popen(
            ["umount", self.part_mask.format(self.device, partition_num)],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, err = unmount_proc.communicate()
        exit_code = unmount_proc.returncode
        if exit_code != 0:
            raise weresync.exception.DeviceError(self.device,
                                                 str(output, "utf-8"))
        # If there are no errors, nothing needs be returned

    def _get_blkid_info(self, partition_num, info_name):
        proc = subprocess.Popen(
            [
                "blkid", self.part_mask.format(self.device, partition_num),
                "-o", "value", "-s", info_name
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = proc.communicate()
        if proc.returncode != 0:
            raise DeviceError(self.device,
                              "Error getting information for partition " +
                              str(partition_num), str(output, "utf-8"))

        return str(output, "utf-8").strip()

    def get_partition_uuid(self, partition_num):
        """Gets the UUID for a given partition. This is *not* the filesystem
        UUID.

        :param partition_num: the number of the partition whose UUID to get.
        :returns: a string containing the partition's UUID"""
        return self._get_blkid_info(partition_num, "UUID")

    def get_part_uuid(self, partition_num):
        """Gets the PARTUUID for a given partition, if it exists. This is
        *not* the filesystem UUID and is different from
        `py:func:~.DeviceManager.get_partition_uuid`.

        :param partition_num: the number of the partition whose PARTUUID to
                              get.
        :returns: a string containing the partition's PARTUUID."""
        return self._get_blkid_info(partition_num, "PARTUUID")

    def get_partition_table_type(self):
        """Gets the type of partition table on the device. Usually "gpt" for
        GPT disks or "msdos" for MBR disks.

        :returns: A string containing the name of the partition table.
        :raises DeviceError: If the parted command has a non-zero return code.
        :raises UnsupportedDeviceError: If the device does not have a
                                        supported partition type."""

        process = subprocess.Popen(
            ["partprobe", "-s", "-d", self.device], stdout=subprocess.PIPE)
        output, error = process.communicate()
        exit_code = process.returncode
        if exit_code != 0:
            raise weresync.exception.DeviceError(self.device,
                                                 "Non-zero exit code",
                                                 str(output, "utf-8"))
        result = str(output, "utf-8")
        for table_type in SUPPORTED_PARTITION_TABLE_TYPES:
            if table_type in result:
                return table_type
        else:
            raise weresync.exception.UnsupportedDeviceError(
                "Partition table "
                "type of {0} not supported by WereSync.".format(self.device))

    def get_drive_size(self):
        """Returns the maximum size of the drive, in sectors."""

        query_proc = subprocess.Popen(
            ["blockdev", "--getsz", self.device], stdout=subprocess.PIPE)
        output, error = query_proc.communicate()
        exit_code = query_proc.returncode
        if exit_code != 0:
            raise weresync.exception.DeviceError(self.device,
                                                 "Non-zero exit code",
                                                 str(output, "utf-8"))

        return int(output)  # should always be valid

    def get_drive_size_bytes(self):
        """Returns the maximum size of the drive, in bytes."""

        query_proc = subprocess.Popen(
            ["blockdev", "--getsize64", self.device], stdout=subprocess.PIPE)
        output, error = query_proc.communicate()
        exit_code = query_proc.returncode
        if exit_code != 0:
            raise weresync.exception.DeviceError(self.device,
                                                 "Non-zero exit code",
                                                 str(output, "utf-8"))

        return int(output)

    def _get_general_info(self, partition_num):
        """Gets general information for the passed partition number, to be
        investigated by other methods. It returns a list with the following
        information in order:

        Filesystem (/dev/sda or such), Size in 512B-blocks, Used, Available,
        Use%, and Mounted Location.
        Gleaned from Linux df command.

        :param partition_num: the partition for which to get the information"""

        mounted_here = False
        try:
            if self.mount_point(partition_num) is None:
                self.mount_partition(partition_num, MOUNT_POINT)
                mounted_here = True

            proc_formal = subprocess.Popen(
                ["df", "--block-size=512"], stdout=subprocess.PIPE)
            proc = subprocess.Popen(
                ["grep", self.part_mask.format(self.device, partition_num)],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                stdin=proc_formal.stdout)
            output, error = proc.communicate()
            exit_code = proc.returncode
            if exit_code >= 2:  # grep returns 2 if an error occurs
                raise weresync.exception.DeviceError(
                    self.device, "Error running grep.",
                    str(output, "utf-8") + str(exit_code))
            elif exit_code == 1:
                raise weresync.exception.DeviceError(self.device,
                                                     "No grep line read", None)

            return [x for x in str(output, "utf-8").split() if x != ""]
            # Output has the following columns:
            # Filesystem, Total-size, Used, Available, Use%, Mount Point

        finally:
            if mounted_here:
                self.unmount_partition(partition_num)

    def get_partition_used(self, partition_num):
        """Returns the space available in a drive in 512B blocks

        :param partition_num: the number of the partition to check"""

        return int(self._get_general_info(partition_num)[2])

    def get_partition_size(self, partition_num):
        """Gets the size of a partition in 512B sectors.

        :param partition_num: An int representing the partition whose size to
                              get.

        :returns: An int representing the number of 512B blocks in the
                  partition
        :raises DeviceError: if the commands getting the size of the partition
                             fail.
        :raises ValueError: if the drive has an unsupported partition table
                            type."""
        table_type = self.get_partition_table_type()
        if table_type == "gpt":
            proc = subprocess.Popen(
                ["sgdisk", self.device, "-p"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = proc.communicate()
            if proc.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.device, "Error getting device information",
                    str(output, "utf-8"))

            for line in str(output, "utf-8").split("\n"):
                if line.strip().startswith(str(partition_num)):
                    words = [x for x in line.split(" ") if x != ""]
                    return int(words[2]) - int(words[1])
                # start sector is the second element in this list, last sector
                # is third element
        elif table_type == "msdos":
            proc = subprocess.Popen(
                [
                    "sfdisk", "-s", self.part_mask.format(self.device,
                                                          partition_num)
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = proc.communicate()
            if proc.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.device,
                    "Error getting partition size for partition {0}".format(
                        partition_num), str(output, "utf-8"))
            return int(output)
        else:
            raise ValueError("Unsupported table type")

    def get_partition_code(self, partition_num):
        """For GPT disks: gets the partition code (as defined by sgdisk) for
        the passed partition number.

        For MBR disks: gets the partition code (as defined by sfdisk) for the
        passed partition number.

        :param partition_num: the number of the partition whose code to find.
        :raises DeviceError: If the command returns a non-zero return code.
        :raises ValueError: If an invalid partition number (one that doesn't
                            exist) is passed.
        :returns: a string containing the partition code for the appropriate
                  disk type."""
        table_type = self.get_partition_table_type()
        if table_type == "gpt":
            proc = subprocess.Popen(
                ["sgdisk", self.device, "-p"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = proc.communicate()
            if proc.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.device,
                    "Error getting device information from sgdisk",
                    str(output, "utf-8"))
            for line in str(output, "utf-8").split("\n"):
                if line.strip().startswith(str(partition_num)):
                    words = line.strip().split()
                    return words[
                        5]  # The code appears in the fifth column, but the
                # size takes up two columns (one for value and one for unit),
                # so the code appears in the sixth column.
        elif table_type == "msdos":
            proc = subprocess.Popen(
                ["fdisk", self.device, "-l"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = proc.communicate()
            if proc.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.device, "Error getting device partition information.",
                    str(output, "utf-8"))

            lines = str(output, "utf-8").split("\n")
            result = list(
                filter(lambda x: x.strip().startswith("Device"), lines))
            header_line = result[0].strip().split()
            for idx, val in enumerate(header_line):
                if val == "Id":
                    code_index = idx
                    break
            else:
                raise ValueError("Unsupported fdisk format.")

            part_name = self.part_mask.format(self.device, partition_num)
            for line in lines:
                if line.strip().startswith(part_name):
                    words = line.strip().split()
                    loc = code_index  # the code is in the 5th column
                    if "*" not in line:
                        loc -= 1
                        # if the partition is not bootable the column for the
                        # boot header is not seperated.

                    return words[loc]

            # No partition found
            raise ValueError("Invalid partition number, no partition found.")

    def get_partition_alignment(self):
        """Returns the number of sectors the drive must be aligned to. For GPT
        disk this is found using sgdisk's output.
        This function is only supported by GPT disks.

        :returns: An integer that represents the partition alignment of the
                  device.
        :raises DeviceError: If the command returns a non-zero return code"""

        table_type = self.get_partition_table_type()
        if table_type == "gpt":
            process = subprocess.Popen(
                ["sgdisk", self.device, "-p"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = process.communicate()
            if process.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.device, "Error finding partition alignment.",
                    str(output, "utf-8"))
            words = [x for x in str(output, "utf-8").split("\n") if x != ""]
            for word in words:
                if word.startswith("Partitions will be aligned on"):
                    result = word.split("Partitions will be aligned on ")
                    # the first value of result will be "", the second value
                    # will be "{alignmentNum}-sector boundaries"
                    value = result[1].split("-")
                    return int(value[0])

            raise weresync.exception.DeviceError(
                self.device, "sgdisk returned abnormal output.")
        elif table_type == "msdos":
            process = subprocess.Popen(
                ["fdisk", self.device, "-l"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = process.communicate()
            if process.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.device, "Error getting partition alignment",
                    str(output, "utf-8"))
            for line in str(output, "utf-8").split("\n"):
                if line.startswith("Sector size"):
                    parts = line.split("Sector size (logical/physical): ")[
                        1].split(" / ")
                    # After the label there is minimum and optimal size
                    # seperated by a slash. Optimal is second
                    logical = int(parts[0].strip().split()[0])
                    physical = int(parts[1].strip().split()[0])
                    return int(physical / logical)

    def get_empty_space(self):
        """Returns the amount of empty space *after* the last partition. This
        does not include space hidden between partitions.

        :returns: An integer representing the number of sectors free after the
                  last partition.
        :raises DeviceError: If the command has a non-zero return code"""
        table_type = self.get_partition_table_type()
        if table_type == "gpt":
            proc = subprocess.Popen(
                ["sgdisk", self.device, "-p"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = proc.communicate()
            if proc.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.device, "Error getting device information.",
                    str(output, "utf-8"))
            result = str(output, "utf-8").split("\n")
            total_sectors = 0
            for line in result:
                if line.startswith("Disk " + self.device):
                    words = line.split(" ")
                    for word in words:
                        try:
                            total_sectors = int(word)
                            break
                        except ValueError:
                            continue

            last_part_info = [
                x for x in result[-2].split(" ") if x != ""
            ]  # the information for the last partition is always the last
            # line in the output. The very last line, however is empty. The
            # last real line is 2 back
            last_sector = int(last_part_info[
                2])  # the third column is the "end" sector column
            return total_sectors - last_sector
        elif table_type == "msdos":
            # other possible table types with throw an UnsupportedDeviceError
            # in the get_partition_table_type() method
            proc = subprocess.Popen(
                ["fdisk", self.device, "-l"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
            output, error = proc.communicate()
            lines = str(output, "utf-8").split("\n")
            for line in lines:
                if "sectors" in line:
                    words = line.split()
                    for word in reversed(words):
                        try:
                            total_sectors = int(word)
                            break
                        except ValueError:
                            continue

                    break

            last_sector = 0
            for val in reversed(lines):
                if val.strip() == "":
                    continue
                vals = val.split()
                loc = 2
                if "*" in val:
                    loc += 1

                try:
                    value = int(vals[loc])
                except ValueError:
                    break  # When this trips, we should be past all the
                # partition descriptions
                if value >= last_sector:
                    last_sector = value

            return total_sectors - value

    def get_partition_file_system(self, part_num):
        """Returns the file system (ext4, ntfs, etc.) of the partition. If a
        partition system that can't be created by this system is found, None
        is return.

        :param part_num: the partition number whose filesystem to get.

        :returns: A string containing the file system type if a type found,
                  otherwise None. If this is a swap partition this returns
                  'swap'"""
        proc = subprocess.Popen(
            [
                "blkid", "-o", "value", "-s", "TYPE",
                self.part_mask.format(self.device, part_num)
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = proc.communicate()
        if proc.returncode != 0 and proc.returncode != 2:
            raise weresync.exception.DeviceError(
                self.part_mask.format(self.device, part_num),
                "Error getting partition file system type.", output)
        result = str(output, "utf-8").strip()
        return result if result in SUPPORTED_FILESYSTEM_TYPES else None

    def set_partition_file_system(self, part_num, system_type):
        """Creates a new file system on the passed partition number. This is
        essentially the same as formatting the partition.
        If the passed partition is currently mounted, this will unmount the
        system, and remount it when finished.

        :param part_num: the partition number to format
        :param system_type: a file system (ex. ntfs) supported by mkfs on the
                            current system."""
        mnt_point = self.mount_point(part_num)
        try:
            if mnt_point is not None:
                self.unmount_partition(part_num)
            if system_type == "swap":
                proc = subprocess.Popen(
                    ["mkswap", self.part_mask.format(self.device, part_num)],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT)
            else:
                proc = subprocess.Popen(
                    [
                        "mkfs", "-t", system_type,
                        self.part_mask.format(self.device, part_num)
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT)
            output, error = proc.communicate()
            if proc.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.part_mask.format(self.device, part_num),
                    "Error creating new file system on partition.",
                    str(output, "utf-8"))
        finally:
            if mnt_point is not None:
                self.mount_partition(part_num, mnt_point)


class LVMDeviceManager(DeviceManager):
    """This is an extension of the DeviceManager class which handles Logical
    Volume groups. All method names referring to \"partition\" in fact refer
    to logical volumes, but the names remain the same for compatibility
    reasons.

    :param source: The name of the logical volume group."""

    def __init__(self, device, partition_mask="{0}/{1}"):
        self.name = device.split("/")[-1]
        self.device = device if not device.endswith("/") else device[0:-1]
        self.device = self.device if self.device.startswith(
            "/dev") else "/dev/" + self.device
        self.part_mask = partition_mask

    def get_partitions(self):
        """Returns the names of the logical volumes in the group, as a list of
        strings."""
        proc = subprocess.Popen(
            ["lvs", "--separator", ":"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = proc.communicate()
        if proc.returncode != 0:
            raise weresync.exception.DeviceError(
                self.device, "Error finding logical volumes.",
                str(output, "utf-8"))
        lines = [
            x.strip().split(":") for x in str(output, "utf-8").split("\n")
            if x.strip() != ""
        ]
        return [x[0] for x in lines[1:] if x[1] == self.name]

    def get_partition_table_type(self):
        return "lvm"

    def _get_drive_size_generic(self, units):
        """Gets the size of a logical volume group in generic units.

        :param units: An indicator for units as defined by the *vgs* command.
        Generally "b" or "s"."""
        proc = subprocess.Popen(
            [
                "vgs", "--units", units, "-o", "size", "--noheadings",
                self.device
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = proc.communicate()
        if proc.returncode != 0:
            raise weresync.exception.DeviceError(
                self.device, "Error finding logical volume size.",
                str(output, "utf-8"))
        return int(
            str(output, "utf-8").strip()[0:-1]
        )  # The last character of the output is a "B" or an "S" and should be

    # removed

    def get_drive_size_bytes(self):
        """Returns the of a logical volume group in bytes."""
        return self._get_drive_size_generic("b")

    def get_drive_size(self):
        """Normally this function returns the drive size in sectors.

        :returns: an integer representing the size of the drive in sectors
        :raises DeviceError: if the command returns a non-zero return code"""
        return self._get_drive_size_generic("s")

    def get_partition_size(self, partition_name):
        """Gets the size, in sectors, of a logical volume.

        :param partition_name: the name of the logical volume whose size to
                               get.

        :returns: an int representing the number of bytes on the logical
                  volume.
        :raise DeviceError: if the commands for getting the size of the
                            partition fail to return a 0 return code."""

        proc = subprocess.Popen(
            [
                "lvdisplay", "-c", self.part_mask.format(self.device,
                                                         partition_name)
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = proc.communicate()
        if proc.returncode != 0:
            raise weresync.exception.DeviceError(
                self.device, "Error getting logical volume size.",
                str(output, "utf-8"))

        return int(str(output, "utf-8").split(":")[6])

    def get_partition_code(self, partition_num):
        """Not valid for LVM drives.

        :raises UnsupportedDeviceError:"""

        raise weresync.exception.UnsupportedDeviceError(
            "Partition codes not applicable to lvm drives.")

    def _get_general_info(self, partition_num):
        """Same method as the _get_general_info method for the main
        DeviceManager class."""
        # This switching masks around is necessary because df sees block
        # devices in mapper (ex. /dev/mapper/example--vg) rather than normal
        # (ex. /dev/example-vg),
        old_name = self.device
        self.device = self.device[self.device.rindex("/") + 1:len(self.device)]
        self.device = self.device.replace("-", "--")
        old_mask = self.part_mask
        self.part_mask = "/dev/mapper/{0}-{1}"
        try:
            result = super()._get_general_info(partition_num)
        finally:
            self.device = old_name
            self.part_mask = old_mask
        return result

    # TODO make this work if needed with LVM drives
    def get_partition_alignment(self):
        """Not valid for LVM drives.

        :raises UnsupportedDeviceError:"""

        raise weresync.exception.UnsupportedDeviceError(
            "Partition codes not applicable to lvm drives.")

    def get_empty_space(self):
        """Gets the amount of empty space left in the logical volume group. Due to
        a lack of strict ordering in lvm, this is considered to be the total
        space, not simply the space after the last partition.

        :returns: An integer representing the number of sectors free in the
                  volume group.
        :raises DeviceError: If the command has a non-zero return value."""
        proc = subprocess.Popen(
            [
                "vgs", "--units", "s", "-o", "free", "--noheadings",
                self.device
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = proc.communicate()
        if proc.returncode != 0:
            raise weresync.exception.DeviceError(
                self.device, "Error reading free space in volume group.",
                str(output, "utf-8"))

        result = str(output, "utf-8").strip()[0:-1]
        # The final character in the results is a unit, in this case "S"
        return int(result)


class DeviceCopier:
    """DeviceCopiers transfer data from a source drive to a target drive.

    :param source: the drive identifier (/dev/sda or such) of the source drive.
    :param target: the drive identifier (/dev/sdb or such) of the target drive.
    :param lvm_source: the name or :py:class:`~.LVMDeviceManager` object
                       representing the source LVM. If None no LVM is copied.
    :param lvm_target: the name or :py:class:`~.LVMDeviceManager` object
                       representing the target LVM. If None no LVM is copied.
    """

    def __init__(self, source, target, lvm_source=None, lvm_target=None):
        self.source = source if isinstance(
            source, DeviceManager) else DeviceManager(source)
        self.target = target if isinstance(
            target, DeviceManager) else DeviceManager(target)
        if lvm_source is not None and lvm_target is not None:
            self.lvm_source = lvm_source if isinstance(
                lvm_source, LVMDeviceManager) else LVMDeviceManager(lvm_source)
            self.lvm_target = lvm_target if isinstance(
                lvm_target, LVMDeviceManager) else LVMDeviceManager(lvm_target)
        else:
            self.lvm_source = None
            self.lvm_target = None
        self.uuid_dict = None

    def get_uuid_dict(self):
        """Generates a dictionary that relates the partitions of the source
        drive to the partitions of the target drive.

        This function requires that both drives have the same number of
        partitions with the same numbers.

        This function caches its result in self.uuid_dict. After changing drive
        uuids, self.uuid_dict should be set to None.

        :returns: A dictionary where the keys are the source drive partition
                  UUIDs, and the values are the corresponding target drive
                  UUIDs."""
        if self.uuid_dict is None:
            uuids = {}
            for i in self.source.get_partitions():
                try:
                    source_uuid = self.source.get_partition_uuid(i)
                except DeviceError as ex:
                    if ex.errors == "":  # blkid returns an empty string if no
                        # UUID found
                        # This can occur for some partition types, like a
                        # microsoft reserved partition
                        continue
                    else:
                        raise ex

                if source_uuid.strip() == "":
                    # If the uuid is empty/nothing, we don't want it
                    continue

                uuids[source_uuid] = self.target.get_partition_uuid(i)
                # This next segment gets the PARTUUID values
                try:
                    source_part_uuid = self.source.get_part_uuid(i)
                except DeviceError as ex:
                    if ex.errors == "":
                        continue

                if source_part_uuid == "":
                    continue

                uuids[source_part_uuid] = self.target.get_part_uuid(i)

            if self.lvm_source is not None:
                print("LVM Source is: " + self.lvm_source.device)
                print("meaningless space")
                source_vg = parse.parse("/dev/{0}", self.lvm_source.device)[0]
                target_vg = parse.parse("/dev/{0}", self.lvm_target.device)[0]
                source_name = source_vg.replace("-", "--")
                target_name = target_vg.replace("-", "--")
                lvm_mask = "{0}-{1}"
                for i in self.lvm_source.get_partitions():
                    uuids[lvm_mask.format(source_name, i)] = lvm_mask.format(
                        target_name, i)

            self.uuid_dict = uuids
            LOGGER.debug("UUID Dict: " + str(uuids))
            return uuids
        else:
            return self.uuid_dict

    def _transfer_gpt(self, difference):
        """Transfers the gpt partition table from the larger source drive to
        the smaller target drive.

        This method is not intended to be called from any method but the
        transfer partition table method below.

        :param difference: the difference between the sizes of the two drives
        :param margin: the amount of margin to give shrunk partitions."""
        partitions = self.source.get_partitions()
        part_alignment = self.target.get_partition_alignment()
        add_args = []
        type_args = []
        difference -= (
            self.source.get_empty_space() - 34
        )  # the empty space in the source does not count against the
        # difference, but we leave the margin just in case
        # It also seems that gpt disks at least have 34 unusable sectors at
        # the end of the empty space, so we remove those from the count
        for i in reversed(partitions):
            drive_size = self.source.get_partition_size(i)
            try:
                part_used = self.source.get_partition_used(i)
            except weresync.exception.DeviceError:
                part_used = drive_size
            space = int(part_alignment * math.floor(
                (drive_size - part_used) / part_alignment))
            part_size = None
            if space > 0 and difference > 0:
                # if the amount of space on the drive is bigger than the
                # difference between the drives
                if space > difference:
                    part_size = int(part_alignment * math.ceil(
                        (drive_size - difference) / part_alignment))
                    difference = 0
                else:
                    part_size = int(part_alignment * math.ceil(part_used /
                                                               part_alignment))
                    difference -= space
            else:
                part_size = int(part_alignment * math.ceil(drive_size /
                                                           part_alignment))
                difference += part_size - drive_size  # this adds or subtracts
                # to the difference based on whether or the aligning changed
                # the size of the partition
            if (i != partitions[-1]):
                # delete_args += ["-d", str(i)]
                add_args = [
                    "-n", "{0}:0:+{1}".format(i, part_size)
                ] + add_args  # part_size is in 512 byte chunks, but gdisk
                # expects a 1 kB based unit if should_break: break
                # this also rounds down, in order to leave more space rather
                # than less
            else:
                # if this is the last partition to be looped, let it occupy
                # maximum space.
                add_args = ["-n", "{0}:0:0".format(i)] + add_args

            type_args += [
                "-t", "{0}:{1}".format(i, self.source.get_partition_code(i))
            ]

        LOGGER.debug(["sgdisk", self.target.device, "-o"] + add_args +
                     type_args)
        clear_process = subprocess.Popen(
            ["sgdisk", self.target.device, "-Z"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        clear_output, _ = clear_process.communicate()
        if clear_process.returncode != 0:
            LOGGER.debug("Clear Process error code: {0}".format(
                clear_process.returncode))
            raise weresync.exception.DeviceError(
                self.target.device, "Error clearing target drive.",
                str(clear_output, "utf-8"))
        copy_process = subprocess.Popen(
            ["sgdisk", self.target.device, "-o"] + add_args + type_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        copy_out, copy_err = copy_process.communicate()
        if copy_process.returncode != 0:
            raise weresync.exception.DeviceError(
                self.target.device, "Error copying partition table to target.",
                str(copy_err, "utf-8"))

        final_proc = subprocess.Popen(
            ["sgdisk", self.target.device, "-G"], stderr=subprocess.STDOUT)
        final_out, final_err = final_proc.communicate()
        if final_proc.returncode != 0:
            raise weresync.exception.DeviceError(
                self.target.device, "Error randomizing GUIDs on target device",
                str(final_out, "utf-8"))

    def _transfer_msdos(self, difference, margin=5):
        """Copies the partition table from a msdos source drive to a target drive.
        :param margin: The percent of original size to leave as a margin in
                       the size of each partition. Is ignored if partition
                       size is 0."""
        for i in self.target.get_partitions():
            if self.target.mount_point(i) is not None:
                self.target.unmount_partition(i)
        current_proc = subprocess.Popen(
            ["sfdisk", "-d", self.source.device],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        partition_table, current_proc_error = current_proc.communicate()
        if current_proc.returncode != 0:
            raise weresync.exception.DeviceError(
                self.source.device, "Error getting partition table backup.",
                str(partition_table, "utf-8"))
        partition_table = str(partition_table, "utf-8")
        partition_listings = [
            x.replace(" ", "") for x in partition_table.split("\n")
            if x.startswith(
                self.source.part_mask.format(self.source.device, ""))
        ]
        for idx, val in enumerate(partition_listings):
            # Standard line of sfdisk -d output:
            # mbr.img1:start=2050,size=1893,Id=83, bootable
            # a new version would have "type" instead of "Id"
            listing = val.split(":")
            pairs = {
                "part": int(
                    parse.parse(
                        self.source.part_mask.format(self.source.device,
                                                     "{0}"), listing[0])[0])
            }
            for i in listing[1].split(","):
                pair = i.split("=")
                if len(pair) != 2:
                    pairs["bootable"] = True
                else:
                    pairs["bootable"] = False
                    if (pair[0] == "Id" or pair[0] == "type"):
                        # Newer versions of sfdisk use type instead of Id
                        # This test allows both versions to be supported
                        pairs["type"] = pair[1]
                        id_key = pair[0]
                        # needed to create the right lines in the
                        # final product
                    else:
                        pairs[pair[0]] = int(pair[1])
            partition_listings[idx] = pairs

        partition_listings.sort(key=lambda x: x["start"])
        move_start_back_by = 0
        final_str = "unit: sectors\n\n"
        current_extended_partition = None

        for i in partition_listings:
            i["start"] -= move_start_back_by
            if current_extended_partition is not None and i[
                    "part"] <= 4:  # Not a logical partition
                current_extended_partition = None

            if i["type"] == "5":  # The partition is an extended partition
                current_extended_partition = i
                continue
            shrink = 0
            try:
                part_num = i["part"]
                drive_used = self.source.get_partition_used(part_num)
                space = math.floor(
                    (i["size"] - drive_used) * (1 - margin / 100))
                if space > 0 and difference > 0:
                    if space >= difference:
                        shrink = difference
                    else:
                        shrink = space
                        difference -= shrink

                    move_start_back_by += shrink
                    i["size"] -= shrink
            except weresync.exception.DeviceError as ex:
                LOGGER.warning("Error reading device.")
                LOGGER.debug("Execption info:", exc_info=sys.exc_info())

            if current_extended_partition is not None:
                current_extended_partition["size"] -= shrink

        partition_listings.sort(key=lambda x: x["part"])
        # We don't know what the name of the id key is, so we have to
        # concaterate it in.
        final_str = "unit: sectors\n\n" + "".join([
            "{val} : start= {start}, size= {size}, {type_key}= {type}{boot}\n".
            format(
                val=self.target.part_mask.format(self.target.device,
                                                 part_line["part"]),
                boot=", bootable" if part_line["bootable"] else "",
                type_key=id_key,
                **part_line) for part_line in partition_listings
        ])

        LOGGER.debug("Proposed partition table:\n" + final_str)

        table_proc = subprocess.Popen(
            ["fdisk", self.target.device],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            universal_newlines=True)
        output, error = table_proc.communicate(input="o\nw\nq")
        if table_proc.returncode != 0:
            raise weresync.exception.CopyError(
                "Could not create new partition table on target device",
                output)
        transfer_proc = subprocess.Popen(
            ["sfdisk", "--force", self.target.device],
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            universal_newlines=True)
        output, error = transfer_proc.communicate(input=final_str)
        if transfer_proc.returncode != 0:
            raise weresync.exception.CopyError(
                "Could not copy partition table to target device.", output)

    def _transfer_lvm(self, difference):
        # In general, partitions and logical volumes (lvs) are synonymous in
        # this method
        for j in self.lvm_target.get_partitions():
            if self.lvm_target.mount_point(j) is not None:
                self.lvm_target.unmount(j)
            LOGGER.debug("Deleting " + j)
            remove_proc = subprocess.Popen(
                ["lvremove", "-f", self.lvm_target.device + "/" + j],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            output, error = remove_proc.communicate()
            if remove_proc.returncode != 0:
                raise weresync.exception.DeviceError(
                    self.lvm_target.device,
                    "Error removing logical volume from target.",
                    str(output, "utf-8"))

        lvs = self.lvm_source.get_partitions()
        difference -= self.lvm_source.get_empty_space()
        for i in lvs:
            drive_size = self.lvm_source.get_partition_size(i)
            try:
                part_used = self.lvm_source.get_partition_used(i)
            except weresync.exception.DeviceError:
                part_used = drive_size

            space = drive_size - part_used
            if space > 0 and difference > 0:
                part_size = None
                if space > difference:
                    part_size = drive_size - difference
                else:
                    part_size = part_used
                    difference -= space
            else:
                part_size = drive_size

            command = [
                "lvcreate", "--size", str(part_size) + "S", "-n", i,
                self.lvm_target.device
            ]
            LOGGER.debug("lvcreate command: " + str(command))
            with tempfile.TemporaryFile() as tmp:
                copy_proc = subprocess.Popen(
                    command, stdout=tmp, stderr=subprocess.STDOUT)
                proc_output, error = copy_proc.communicate()
                tmp.seek(0)
                output = tmp.read()
                LOGGER.debug("Output for " + i + ": " + str(output, "utf-8"))
                if copy_proc.returncode != 0:
                    raise weresync.exception.DeviceError(
                        self.lvm_target.device,
                        "Error creating new logical volume.",
                        str(output, "utf-8"))

    def format_partitions(self, ignore_errors=True, callback=None, lvm=False):
        """Goes through each partition in the source drive and formats the
        corresponding partition in the target drive to the same thing.

        :param ignore_errrors: whether or not errors (such as a file-system
                               not recongized by mkfs) should be ignored. If
                               True such errors will be logged. If false the
                               exceptions will be propogated. Defaults to True.
        :param callback: If not None, this function should be a function that
                         accepts a float between 0 and 1 to update the
                         progress.
        :param lvm: Whether or not to use the lvm managers."""
        if lvm:
            source_manager = self.lvm_source
            target_manager = self.lvm_target
        else:
            source_manager = self.source
            target_manager = self.target

        partitions = source_manager.get_partitions()
        for i in partitions:
            try:
                part_type = source_manager.get_partition_file_system(i)
                if callback is not None:
                    drive_size = target_manager.get_drive_size()
                    complete = 0.0
                if part_type is not None:
                    target_manager.set_partition_file_system(i, part_type)
                    if callback is not None:
                        part_size = target_manager.get_partition_size(i)
                        complete += part_size / drive_size
                        LOGGER.debug("Callback:\nDrive Size: {0}\n"
                                     "Part Size: {1}\n"
                                     "Complete: {2}".format(
                                         drive_size, part_size, complete))
                        callback(complete)
                else:
                    LOGGER.warning(
                        "Invalid filesystem type found. Partition {0} not "
                        "formatted.".format(i))
            except weresync.exception.DeviceError as exe:
                if ignore_errors:
                    logging.getLogger("weresync.device").warning(
                        "Creating filesystem for {0} encountered errors. "
                        "Partition type: {1}. Skipped.".format(
                            target_manager.part_mask.format(
                                target_manager.device, i), part_type),
                        exc_info=sys.exc_info())
                    logging.getLogger("weresync.device").debug(
                        "Error making file system.", exc_info=sys.exc_info())
                else:
                    raise exe

    def transfer_partition_table(self, resize=True, callback=None):
        """Transfers the partition table from one drive to another. Afterwards,
        it formats the partitions on the target drive to be the same as those
        on the source drive.

        :param resize: if true (default) then the program will attempt to
                       resize the partition tables to fit on a smaller drive.
                       This will not expand partition tables at any time.
        :param callback: If not none, a function to update progress completed.
                         See py:func:`~.format_partitions()`"""

        source_size = self.source.get_drive_size()
        target_size = self.target.get_drive_size()

        source_type = self.source.get_partition_table_type()
        if target_size < source_size and not resize:
            raise weresync.exception.CopyError(
                "Target device smaller than source device and resize set to "
                "false.")
        elif source_type == "gpt":
            self._transfer_gpt(source_size - target_size)
        elif source_type == "msdos":
            self._transfer_msdos(source_size - target_size)

        for i in self.target.get_partitions():
            if self.target.mount_point(i) is not None:
                self.target.unmount_partition(i)

        if callback is not None:
            callback(0.3)

        # the block devices still won't be updated unless the following
        # command is called.
        proc = subprocess.Popen(
            ["partprobe", self.target.device],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        output, error = proc.communicate()
        if proc.returncode != 0:
            raise weresync.exception.DeviceError(
                self.target.device, "Error reloading partition mappings.",
                str(output, "utf-8"))

        # This weights the progress so 30% comes from creating partition and
        # 70% from formatting.
        progress = lambda prog: (callback(0.3 + prog * 0.7) if callback  # noqa
                                 is not None else None)
        self.format_partitions(callback=progress)
        # If it's done, it's done
        if callback is not None:
            callback(1.0)
        self.uuid_dict = None  # uuids will have been changed.

    def transfer_lvm_partition(self, resize=True, callback=False):
        if self.lvm_source is not None:
            lvm_source_size = self.lvm_source.get_drive_size()
            lvm_target_size = self.lvm_target.get_drive_size()
            self._transfer_lvm(lvm_source_size - lvm_target_size)
            self.format_partitions(callback=callback, lvm=True)

    def partitions_valid(self, lvm=False):
        """Tests if the partitions on the target drive can support copying
        files from the source drive.

        :returns: True if no errors found.
        :raises: a :py:class:`~weresync.exception.CopyError` if any part
                 invalid."""
        if lvm:
            source_manager = self.lvm_source
            target_manager = self.lvm_target
        else:
            source_manager = self.source
            target_manager = self.target

        source_parts = source_manager. get_partitions()
        try:
            target_parts = target_manager.get_partitions()
        except DeviceError as ex:
            raise weresync.exception.CopyError(
                "Target partitions cannot be found. Invalid.")
        if source_parts != target_parts:
            raise weresync.exception.CopyError(
                "Partition count on two drives different. Invalid.")
        for i in source_parts:
            if source_manager.get_partition_file_system(
                    i) != target_manager.get_partition_file_system(i):
                raise weresync.exception.CopyError(
                    "File system type for partition {0} does not match. "
                    "Invalid.".format(i))
            try:
                if source_manager.get_partition_used(
                        i) > target_manager.get_partition_size(i):
                    raise weresync.exception.CopyError(
                        "Information on partition {0} cannot fit on "
                        "corresponding partition on target drive.".format(i))
            except weresync.exception.DeviceError as ex:
                if "mount" in str(ex) or "swapspace" in str(
                        ex
                ):  # the partition couldn't be mounted because it isn't a
                    # mountable partition (maybe boot sector or swap)
                    LOGGER.debug(
                        "Partition {0} couldn't be mounted. Bad FS type".
                        format(i),
                        exc_info=sys.exc_info)
                else:
                    raise ex

        return True

    def _copy_fstab(self, mnt_source, mnt_target, excluded_partitions=[],
                    lvm=False):
        """Updates files in /etc/fstab to be bootable on the target drive.

        :param mnt_source: the directory where source partitions should be
                           mounted.
        :param mnt_target: the directory where target partitions should be
                           mounted.
        :param excluded_partitions: partitions not to search. Defaults to
                                    empty"""
        if lvm:
            source_manager = self.lvm_source
            target_manager = self.lvm_target
        else:
            source_manager = self.source
            target_manager = self.target

        for i in source_manager.get_partitions():
            source_mounted = False
            target_mounted = False
            try:
                if i not in excluded_partitions:
                    source_loc = source_manager.mount_point(i)
                    if source_loc is None:
                        try:
                            source_manager.mount_partition(i, mnt_source)
                            source_mounted = True
                            source_loc = mnt_source
                        except weresync.exception.DeviceError as ex:
                            if "mount" in str(ex):
                                LOGGER.debug(
                                    "Failed to mount partition. Info:\n",
                                    exc_info=sys.exc_info())
                                continue
                            else:
                                raise ex
                    source_fstab_path = source_loc + (
                        "/"
                        if not source_loc.endswith("/") else "") + "etc/fstab"
                    if os.path.exists(source_fstab_path):
                        target_loc = target_manager.mount_point(i)
                        if target_loc is None:
                            target_manager.mount_partition(i, mnt_target)
                            target_mounted = True
                            target_loc = mnt_target
                        with open(source_fstab_path) as source_fstab, open(
                                target_loc + ("/"
                                              if not target_loc.endswith("/")
                                              else "") + "etc/fstab",
                                "w") as target_fstab:
                            target_fstab.write(
                                "# This file is generated by WereSync. All"
                                " comments have been copied, but they have not"
                                " been parsed.\n# Any reference to"
                                " identifiers during installation may be"
                                " inaccurate.\n\n")
                            for line in source_fstab.readlines():
                                stripLine = line.strip()
                                if stripLine == "" or stripLine.startswith(
                                        "#"):
                                    target_fstab.write(line)
                                    continue

                                words = stripLine.split()

                                if words[0].startswith("UUID") or words[
                                        0].startswith("LABEL"):

                                    if words[0].startswith("UUID"):
                                        blkid_arg = "-U"
                                    elif words[0].startswith("LABEL"):
                                        blkid_arg = "-L"

                                    identifier = words[0].split("=")[
                                        1]  # First argument of the space
                                    # separated list is the identifier,
                                    # which is split by a equals sign, and the
                                    # second value is the identifier
                                    proc = subprocess.Popen(
                                        ["blkid", blkid_arg, identifier],
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT)
                                    output, error = proc.communicate()
                                    if proc.returncode == 2:
                                        raise weresync.exception.DeviceError(
                                            self.source.device,
                                            "Could not find block name of "
                                            "device with id: {0}".format(
                                                identifier),
                                            str(output, "utf-8"))
                                    elif proc.returncode != 0:
                                        raise weresync.exception.DeviceError(
                                            source_manager.device,
                                            "Error finding device error for "
                                            "device with id: {0}".format(
                                                identifier),
                                            str(output, "utf-8"))
                                    # It figures out the value of a
                                    # placeholder based on context. However,
                                    # the part_masks rarely have enough context
                                    # So we format the part_mask so that the
                                    # first placeholder is the partition
                                    # number, and the device name is inserted
                                    # (comes out to something like
                                    # "/dev/nbd0p{0}").
                                    # Then it can figure it out.
                                    out = str(output, "utf-8").strip()
                                    if (self.lvm_source is not None
                                        and self.lvm_source.device in out):
                                        source = self.lvm_source
                                        target = self.lvm_target
                                    else:
                                        source = self.source
                                        target = self.target
                                    result = parse.parse(source.part_mask.
                                                         format(source.device,
                                                                "{0}"), out)[0]
                                    # the first element contains the
                                    # number
                                    uuid_proc = subprocess.Popen(
                                        [
                                            "blkid",
                                            source.part_mask.format(
                                                target.device, result)
                                        ],
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT)
                                    uuid_output, uuid_error = (
                                        uuid_proc.communicate())
                                    if uuid_proc.returncode == 2:
                                        raise weresync.exception.DeviceError(
                                            target.device,
                                            "Could not find device {0}".format(
                                                target.part_mask.
                                                format(
                                                    target.device,
                                                    result)),
                                            str(uuid_output, "utf-8"))
                                    elif uuid_proc.returncode != 0:
                                        raise weresync.exception.DeviceError(
                                            target.device,
                                            "Error finding uuid for device "
                                            "{0}".format(
                                                target.part_mask.
                                                format(
                                                    target.device,
                                                    result)),
                                            str(uuid_output, "utf-8"))
                                    ids = str(uuid_output, "utf-8").split()
                                    for val in ids:
                                        if val.startswith("UUID"):
                                            words[0] = val.replace('"', '')
                                            # Have to remove the double
                                            # quotes from blkid's output.
                                            break
                                elif self.lvm_source is not None:
                                    words[0] = multireplace(
                                        words[0], self.get_uuid_dict())
                                target_fstab.write(" ".join(words) + "\n")
            finally:
                if source_mounted:
                    source_manager.unmount_partition(i)
                if target_mounted:
                    source_manager.unmount_partition(i)

    def _copy_files(self,
                    mnt_source,
                    mnt_target,
                    excluded_partitions,
                    ignore_failures,
                    rsync_args,
                    callback,
                    lvm=False):
        """This is an internal method used for copying files. See the
        main `copy_files` method for documentation."""
        if lvm:
            source_manager = self.lvm_source
            target_manager = self.lvm_target
        else:
            source_manager = self.source
            target_manager = self.target
        for i in source_manager.get_partitions():
            source_mounted = False
            target_mounted = False
            try:

                if i in excluded_partitions:
                    continue

                source_loc = source_manager.mount_point(i)
                if source_loc is None:
                    source_manager.mount_partition(i, mnt_source)
                    source_mounted = True
                    source_loc = mnt_source
                target_loc = target_manager.mount_point(i)
                if target_loc is None:
                    target_manager.mount_partition(i, mnt_target)
                    target_mounted = True
                    target_loc = mnt_target

                LOGGER.info("Starting rsync process for partition {0}.".format(
                    source_manager.device))
                command_args = ["rsync"] + shlex.split(rsync_args) + [
                    '--exclude=' + x + ''
                    for x in [
                        "/dev/*", "/proc/*", "/sys/*", "/tmp/*", "/run/*",
                        "/mnt/*", "/media/*", "/lost+found", "/home/*/.gvfs"
                    ]
                ] + [
                    source_loc +
                    ("/" if not source_loc.endswith("/") else ""), target_loc
                ]
                if callback is not None:
                    command_args += ["--info=progress2"]
                print("Copying partition " + str(i))
                LOGGER.debug("Arguments = " + " ".join(command_args))

                def run_proc():
                    with subprocess.Popen(
                            command_args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE) as proc:
                        buf = bytearray()
                        while True:
                            byt = proc.stdout.read(1)
                            if byt == b"":
                                break
                            elif byt == b"\r":
                                yield buf.decode()
                                buf = bytearray()
                            else:
                                buf += byt

                        LOGGER.debug("Errors for partition {0}:\n".format(i) +
                                     proc.stderr.read().decode())

                if callback is not None:
                    for val in run_proc():
                        vals = val.split()
                        if len(vals) >= 2 and vals[1].endswith("%"):
                            try:
                                float_val = float(vals[1].strip("%")) / 100
                            except ValueError:
                                continue
                            callback(i, float_val)

                    LOGGER.debug("Setting to finished")
                    callback(i, 1.0)
                else:
                    for val in run_proc():
                        # If you don't loop through the generator values, the
                        # rsync process doesn't run properly.
                        pass

            except weresync.exception.DeviceError as exe:
                if ignore_failures:
                    LOGGER.warning(
                        "Error copying data for partition {0} from device {1} "
                        "to {2}.".format(i, source_manager.device,
                                         target_manager.device))
                    LOGGER.debug("Error info.", exc_info=sys.exc_info())
                    if callback is not None:
                        callback(i, -1.0)
                else:
                    raise exe
            finally:
                if source_mounted:
                    source_manager.unmount_partition(i)
                if target_mounted:
                    target_manager.unmount_partition(i)
        print(_("Finished copying files."))

    def copy_files(self,
                   mnt_source,
                   mnt_target,
                   excluded_partitions=[],
                   ignore_failures=True,
                   rsync_args=DEFAULT_RSYNC_ARGS,
                   callback=None):
        """Copies all files from source to target drive, doing one partition
        at a time. This assumes that the two drives have equivalent partition
        mappings, i.e. that the data on partition 1 of the source drive should
        be on partition 1 of the target drive.

        :param mnt_source: The directory to mount partitions from the source
                           drive on.
        :param mnt_target: The directory to mount partitions from the target
                           drive on.
        :param excluded_partitions: A list containing the partitions to not
                                    copy. Defaults to empty.
        :param ignore_failures: If True, errors encountered for a partition
                                will not cause the function to exit, but we
                                instead cause a warning to be logged. Defaults
                                to true.
        :param callaback: If not None, a function with the signature
                          ``callback(int, float)``, where the int represents
                          partition number and float represents progress. If
                          an error occurs, the float will be negative. If the
                          float should pulse, it wil return True."""
        self._copy_files(mnt_source, mnt_target, excluded_partitions,
                         ignore_failures, rsync_args, callback, lvm=False)
        if self.lvm_source is not None:
            self._copy_files(mnt_source, mnt_target, excluded_partitions,
                             ignore_failures, rsync_args, callback, lvm=True)

    def make_bootable(self,
                      plugin_name,
                      source_mnt,
                      target_mnt,
                      excluded_partitions=[],
                      root_partition=None,
                      boot_partition=None,
                      efi_partition=None,
                      callback=None):
        """Calls the appropriate plugin to make the target drive bootable.

        :param plugin_name: the name, not pretty name, of the plugin to use.
                            If None, so bootloading occurs, other than fstab
                            copying.
        :param source_mnt: a string representing the directory where
                           partitions from the source drive may be mounted.
        :param target_mnt: a string representing the directory where
                           partitions from the target drive may be mounted.
        :param copier: an instance of :py:class:`~weresync.device.DeviceCopier`
                       which represents the source and target drives.
        :param excluded_partitions: these partitions should not be searched or
                                    included in the boot installation.
        :param boot_partition: this is the partition that should be mounted on
                               /boot of the root_partition.
        :param root_partition: this is the root partition of the drive, where
                               the bootloader should be installed.
        :param efi_partition: this is the partition of the Efi System
                              Partition. Should be None if not a UEFI system.
        :param callback: a function which takes a single argument declaring
                         whether or not the bootloader was successfully
                         installed.
        """
        LOGGER.info("Using plugin: " + plugin_name)
        if callback is not None:
            callback(False)
        try:
            if plugin_name is not None:
                import weresync.plugins as plugins
                manager = plugins.get_manager()
                manager.collectPlugins()
                full_name = "weresync_" + plugin_name
                pluginInfo = manager.getPluginByName(full_name, "bootloader")
                if pluginInfo is None:
                    raise PluginNotFoundError("No such plugin {0}".format(
                        full_name))
                manager.activatePluginByName(full_name, "bootloader")
                plugin = pluginInfo.plugin_object
            else:
                plugin = None

            try:
                self._copy_fstab(source_mnt, target_mnt, excluded_partitions)
            except DeviceError as ex:
                LOGGER.warning("Error copying fstab. Continuing anyway.")
                LOGGER.debug("Info: ", exc_info=sys.exc_info())

            if self.lvm_source is not None:
                try:
                    self._copy_fstab(source_mnt, target_mnt,
                                     excluded_partitions, lvm=True)
                except DeviceError as ex:
                    LOGGER.warning("Error copying fstab on LVM. Continuing"
                                   " anyway.")
                    LOGGER.debug("Info: ", exc_info=sys.exc_info())

            if plugin is not None:
                plugin.install_bootloader(source_mnt, target_mnt, self,
                                          excluded_partitions, boot_partition,
                                          root_partition, efi_partition)
                manager.deactivatePluginByName(plugin_name, "bootloader")
            else:
                LOGGER.warning("No bootloader plugin specified. Not installing"
                               "bootloader.")
            if callback is not None:
                callback(True)
        except DeviceError as ex:
            LOGGER.warning("Error copying bootloader.")
            LOGGER.debug("Info: ", exc_info=sys.exc_info())



# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Connects the interface applications (cli and gui) with the backend daemon.

See also :mod:`weresync.daemon.daemon`"""
from pydbus import SystemBus
from gi.repository import GLib
import threading
import gi.repository.GLib

try:
    bus = SystemBus()
    drive_copier = bus.get("net.manilas.weresync.DriveCopier")
    copy_drive = drive_copier.CopyDrive
    """A function which connects to :meth:`weresync.daemon.device.copy_drive`:.
    For arguments see that function's documentation. If False, WereSync failed
    to connect to dbus."""
    error = None
except gi.repository.GLib.Error as e:
    error = e
    """The error created by connecting to dbus. If no error, then None."""
    copy_drive = False


def _unthreaded_subscribe_to_signals(
        partition_status_callback, copy_status_callback, boot_status_callback):
    drive_copier.PartitionStatus.connect(partition_status_callback)
    drive_copier.CopyStatus.connect(copy_status_callback)
    drive_copier.BootStatus.connect(boot_status_callback)
    loop = GLib.MainLoop()
    loop.run()


def subscribe_to_signals(partition_status_callback, copy_status_callback,
                         boot_status_callback):
    """Subscribes to the weresync callbacks to get information from the daemon.

    :param partition_status_callback: A function which takes a float between 0
                                      and 1 which represents the progress on
                                      formatting partitions.
    :param copy_status_callback: A function which takes an integer representing
                                 the partition being copied and a float between
                                 0 and 1 representing the progress on the copy.
    :param boot_status_callback: A function which takes a boolean. If passed
                                 False the bootloader has not finished
                                 installing. If True it has."""
    thread = threading.Thread(
        target=_unthreaded_subscribe_to_signals,
        args=(partition_status_callback, copy_status_callback,
              boot_status_callback))
    thread.daemon = True
    thread.start()



# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Installs the syslinux bootloader. This bootloader plugin requires the
"root partition" option to be defined.

Setups using syslinux's altmbr setup are currently not supported.

This plugin assumes that the boot folder is in /boot/syslinux. If this is
not the case a symbolic link should be created before WereSync is run.

This plugin depends on Extlinux being installed.

On UEFI systems this simply runs the UUID Copy plugin."""
from weresync.plugins import IBootPlugin
import weresync.plugins as plugins
from weresync.exception import CopyError, DeviceError
import subprocess


class SyslinuxPlugin(IBootPlugin):

    def __init__(self):
        super().__init__("syslinux", "Syslinux")

    def get_help(self):
        return __doc__

    def install_bootloader(self, source_mnt, target_mnt, copier,
                           excluded_partitions=[],
                           boot_partition=None, root_partition=None,
                           efi_partition=None):
        if root_partition is None and boot_partition is None:
            boot_part = plugins.search_for_boot_part(target_mnt, copier.target,
                                                     "syslinux",
                                                     excluded_partitions)
            if boot_part is None and copier.lvm_source is not None:
                boot_part = plugins.search_for_boot_part(target_mnt,
                                                         copier.lvm_source,
                                                         "syslinux",
                                                         excluded_partitions)
            if boot_part is None:
                raise CopyError("Could not find partition with 'syslinux' "
                                "folder on device {0}.".format(copier.target.
                                                               device))
            boot_partition = boot_part
        elif boot_partition is None:
            boot_partition = root_partition

        plugins.translate_uuid(copier, boot_partition, "/boot", target_mnt)

        if efi_partition is not None:
            plugins.translate_uuid(copier, efi_partition, "/", target_mnt)
        else:
            if root_partition is None:
                raise CopyError("The syslinux bootloader plugin requires that"
                                " the root partition be defined. The UUIDs of"
                                " the /boot folder have been updated.")
            try:
                mounted_here = False
                mount_point = copier.target.mount_point(root_partition)
                if mount_point is None:
                    copier.target.mount_partition(root_partition, target_mnt)
                    mount_point = target_mnt
                    mounted_here = True
                mount_point += "/" if not target_mnt.endswith("/") else ""
                extlinux_proc = subprocess.Popen(["extlinux", "--install",
                                                  mount_point + "boot/syslinux"
                                                  ],
                                                 stdout=subprocess.PIPE,
                                                 stderr=subprocess.STDOUT)
                extlinux_output, _ = extlinux_proc.communicate()
                if extlinux_proc.returncode != 0:
                    raise CopyError("Error installing bootloader on target "
                                    "drive.",
                                    extlinux_output)
                table_type = copier.target.get_partition_table_type()
                if table_type == "msdos":
                    bios_proc = subprocess.Popen(["dd", "bs=440", "count=1",
                                                  "if={0}usr/lib/syslinux"
                                                  "/bios/mbr.bin".format(
                                                      mount_point),
                                                  "of=" + copier.target.device
                                                  ], stdout=subprocess.PIPE,
                                                 stderr=subprocess.STDOUT)
                    output, error = bios_proc.communicate()
                    if bios_proc.returncode != 0:
                        raise DeviceError(copier.target.device,
                                          "Error installing bios to drive.",
                                          output)
                elif table_type == "gpt":
                    attribute_proc = subprocess.Popen(["sgdisk",
                                                       copier.target.device,
                                                       "--attributes=1:set:2"],
                                                      stdout=subprocess.PIPE,
                                                      stderr=subprocess.STDOUT)
                    output, error = attribute_proc.communicate()
                    if attribute_proc.returncode != 0:
                        raise DeviceError(copier.target.device,
                                          "Error enabling boot of partition.",
                                          output)
                    bios_proc = subprocess.Popen(["dd", "conv=notrunc",
                                                  "count=1", "if={0}usr/lib"
                                                  "/syslinux/bios/gptmbr.bin".
                                                  format(mount_point),
                                                  "of=" + copier.target.device
                                                  ], stdout=subprocess.PIPE,
                                                 stderr=subprocess.STDOUT)
                    b_output, b_error = bios_proc.communicate()
                    if bios_proc.returncode != 0:
                        raise DeviceError(copier.target.device,
                                          "Error install MBR to drive.",
                                          b_output)
            finally:
                if mounted_here:
                    copier.target.unmount_partition(root_partition)



# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,

# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package contains code for bootloader plugins.

Bootloader plugin is any class extending
:py:class:`~weresync.plugins.IBootPlugin` inside a file name following the
regex pattern "^weresync_.*\.py$". These plugins can be installed to the
site-packages directory (for example, using pip) or to
/usr/local/weresync/plugins.

For an example plugin see the `GrubPlugin <https://github.com/DonyorM/weresync/blob/master/src/weresync/plugins/weresync_grub2.py>`_.""" # noqa

from yapsy.PluginManager import PluginManager
from yapsy.PluginFileLocator import (PluginFileAnalyzerMathingRegex,
                                     PluginFileLocator)
from yapsy.IPlugin import IPlugin
from distutils.sysconfig import get_python_lib
import os
import os.path
import weresync.daemon.device as device
from weresync.exception import DeviceError
import sys
import logging

LOGGER = logging.getLogger(__name__)


def translate_uuid(copier, partition, path, target_mnt):
    """Translates all uuids of the files in the given partition at path.
        This will not affect files which are not UTF-8 or ASCII, and it will
        not affect files which are greater than 200 MB.

        :param copier: the object with the DeviceManager instances.
        :param partition: the partition number of the partition to translate.
        :param path: the path of the file to change relative to the mount of
                     the partition. Should start with "/".
        :param target_mnt: the path to the folder where the partitions should
                           be mounted."""

    mounted_here = False
    try:
        mount_point = copier.target.mount_point(partition)
        if mount_point is None:
            copier.target.mount_partition(partition, target_mnt)
            mount_point = target_mnt
            mounted_here = True

        for dname, dirs, files in os.walk(mount_point + path):
            for fname in files:
                # This if block seeks to avoid opening huge files since
                # they are unlikely to be config files like we are looking
                # for.
                fpath = os.path.join(dname, fname)
                if (os.path.getsize(fpath)) / 1000000 > 200:
                    continue

                try:
                    with open(fpath) as file:
                        text = file.read()
                except UnicodeDecodeError as ex:
                    continue

                uuid_dict = copier.get_uuid_dict()
                text = device.multireplace(text, uuid_dict)
                with open(fpath, "w") as f:
                    f.write(text)
    finally:
        if mounted_here:
            copier.target.unmount_partition(partition)


def mount_partition(manager, lvm_manager, part, mount_point):
    """Mounts a partition and figures out whether or not the partition
        is in the LVM drive. It assumes a numerical partition is not a
        logical volume.

        :param manager: a :py:class:`~weresync.device.DeviceManager` object
                        representing a possible mount.
        :param lvm_manager: a :py:class:`~weresync.device.LVMDeviceManager`
                            object representing a possible host for the mount.
        :param part: the name or number of the partition to mount.
        :param mount_point: the location to mount the partition."""

    try:
        part_num = int(part)
        manager.mount_partition(part_num, mount_point)
        return
    except ValueError:
        pass

    lvm_manager.mount_partition(part, mount_point)


def search_for_boot_part(
        target_mnt,
        target_manager,
        search_folder,
        exlcuded_partitions=[],
):
    """Finds the partition that is the boot partition, by searching for
        a specific folder name. The first partition that contains this name
        or /boot/<name> will be returned.

        :param target_mnt: the folder to mount partitions
        :param target_manager: The :py:class:`~weresync.device.DeviceManager`
                               class representing the drive to search.
        :param search_folder: The name of the folder to search for
        :param excluded_partitions: A list containing a list of partitions
                                    which should not be searched."""
    for i in target_manager.get_partitions():
        if i in exlcuded_partitions:
            continue
        try:
            mounted_here = False
            mount_point = target_manager.mount_point(i)
            if mount_point is None:
                target_manager.mount_partition(i, target_mnt)
                mount_point = target_mnt
                mounted_here = True
            mount_point += "/" if not mount_point.endswith("/") else ""
            if (os.path.exists(mount_point + "boot/" + search_folder)
                    or os.path.exists(mount_point + search_folder)):
                return i
        except DeviceError as ex:
            LOGGER.warning("Could not mount partition {0}. "
                           "Assumed to not be the partition grub "
                           "is on.".format(i))
            LOGGER.debug("Error info:\n", exc_info=sys.exc_info())
        finally:
            try:
                if mounted_here:
                    target_manager.unmount_partition(i)
            except DeviceError as ex:
                LOGGER.warning("Error unmounting partition " + i)
                LOGGER.debug("Error info:\n", exc_info=sys.exc_info())
    else:  # No partition found
        return None


class IBootPlugin(IPlugin):
    """An interface class for bootloader plugins. Plugins implementing this class
    must implement the :py:func:`~.IBootPlugin.install_bootloader` method.

    The name of a plugin should simply be its filename, without the
    "weresync\_" prefix or a file extension. So "weresync_grub2.py"'s name
    would be "grub2".

    :param prettyName: a human readable name for display. Can contain any
                       character.
    :param name: A unique identifying name that should be easy to type on
                 a terminal. Used by users to tell WereSync which plugin
                 to use. See above for exact definition."""

    def __init__(self, name, prettyName=None):
        self.name = name
        if prettyName is None:
            self.prettyName = name
        else:
            self.prettyName = prettyName

    def activate(self):
        """Called at plugin activation,right before bootloader is installed.

        This will be called before /etc/fstab has been updated."""
        pass

    def deactivate(self):
        """Called at plugin deactivation, after bootloader has been
        installed."""
        pass

    def install_bootloader(self,
                           source_mnt,
                           target_mnt,
                           copier,
                           excluded_partitions=[],
                           boot_partition=None,
                           root_partition=None,
                           efi_partition=None):
        """Called to make the drive bootable.

        This will be called after /etc/fstab has been updated.

        :param source_mnt: a string representing the directory where
                           partitions from the source drive may be mounted.
        :param target_mnt: a string representing the directory where
                           partitions from the target drive may be mounted.
        :param copier: an instance of :py:class:`~weresync.device.DeviceCopier`
                       which represents the source and target drives.
        :param excluded_partitions: these partitions should not be searched or
                                    included in the boot installation.
        :param boot_partition: this is the partition that should be mounted on
                               /boot of the root_partition.
        :param root_partition: this is the root partition of the drive, where
                               the bootloader should be installed.
        :param efi_partition: this is the partition of the Efi System
                              Partition. Should be None if not a UEFI system.
        :raises DeviceError: if a bootloader installation command has an error.
        """
        pass

    def get_help(self):
        """Returns the help message for this plugin.

        It is optional to override this.

        :returns: a string representing the help message for this plugin."""
        return "Installs the {0} bootloader.".format(self.prettyName)


dirs = [
    "/usr/local/weresync/plugins",
    os.path.dirname(__file__),
    get_python_lib()
]

regex_analyzer = PluginFileAnalyzerMathingRegex("regex", "^weresync_.*\.py$")
locator = PluginFileLocator([regex_analyzer])

__manager = PluginManager(
    categories_filter={"bootloader": IBootPlugin},
    directories_list=dirs,
    plugin_locator=locator)


def get_manager():
    """Returns the PluginManager for this instance of WereSync"""
    return __manager



# Copyright 2016 Daniel Manila
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module runs the GUI for WereSync."""

import weresync.plugins as plugins
import weresync.utils as utils
import weresync.daemon.device as device
from weresync.exception import InvalidVersionError
import subprocess
import gi
import sys
import os
import logging
import logging.handlers
import threading
gi.require_version("Gtk", '3.0')
from gi.repository import Gtk, GLib, GObject  # noqa

connected = False

try:
    import weresync.interface.dbus_client as dbus_client
    connected = True
except GLib.GError as ex:
    pass


LOGGER = logging.getLogger(__name__)

DEFAULT_HORIZONTAL_PADDING = 5
DEFAULT_VERTICAL_PADDING = 3


class NumberEntry(Gtk.Entry):
    def __init__(self, allowed="", *args, **kargs):
        """An entry that only allows numbers and certain other characters.

        :param allowed: The other characters allowed by this entry in an
                        unseperated string, ex. '., ' to allow periods, commas,
                        and spaces. Defaults to none."""
        Gtk.Entry.__init__(self, *args, **kargs)
        self.allowed = allowed
        self.connect('changed', self.on_changed)

    def on_changed(self, *args):
        text = self.get_text()
        self.set_text("".join(
            [i for i in text if i in "0123456789" + self.allowed]))


def set_margin(widget,
               right=DEFAULT_HORIZONTAL_PADDING,
               left=DEFAULT_HORIZONTAL_PADDING,
               top=DEFAULT_VERTICAL_PADDING,
               bottom=DEFAULT_VERTICAL_PADDING):
    widget.set_margin_right(right)
    widget.set_margin_left(left)
    widget.set_margin_top(top)
    widget.set_margin_bottom(bottom)


def create_help_box(parent, text, title=""):
    help = Gtk.Label(
        halign=Gtk.Align.START,
        xpad=DEFAULT_HORIZONTAL_PADDING,
        ypad=DEFAULT_VERTICAL_PADDING)
    help.set_markup(_("<a href=\"#\">What's this?</a>"))

    def help_click(*args):
        dialog = Gtk.MessageDialog(parent, 0, Gtk.MessageType.INFO,
                                   Gtk.ButtonsType.OK, title)
        dialog.format_secondary_text(text)
        dialog.set_default_size(parent.get_size()[0], -1)
        dialog.run()
        dialog.destroy()
        return True

    help.connect("activate-link", help_click)
    return help


def generate_drive_list():
    proc = subprocess.Popen(
        ["lsblk", "-dnoNAME"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)
    output, _ = proc.communicate()
    if proc.returncode != 0:
        LOGGER.critical("Error reading block list.\n" + output)
    device_list = [
        "/dev/" + x.strip() for x in str(output, "utf-8").split("\n")
        if x.strip() != ""
    ]
    return device_list


def generate_vg_list():
    try:
        lvm_proc = subprocess.Popen(
            ["vgs", "-o", "name", "--noheadings"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        lvm_output, _ = lvm_proc.communicate()
        out = str(lvm_output, "utf-8").split("\n")
        if lvm_proc.returncode != 0:
            LOGGER.critical("Error reading volume group list.\n" +
                            " ".join(out))
        if "No volume groups found" in out:
            lvm_list = ["No volume groups found"]
        else:
            lvm_list = ["/dev/" + x.strip() for x in out if x.strip() != ""]
    except FileNotFoundError as ex:
        # Probably means LVM is not installed on the system, which is no big
        # deal. We'll just log the exception and move on
        LOGGER.debug("File not found info: ", exc_info=sys.exc_info())
        lvm_list = []  # This variable needs to be defined

    return lvm_list


def get_resource(resource):
    dir = os.path.dirname(__file__)
    rel_resource_path = os.path.join(dir, "../resources", resource)
    return os.path.abspath(rel_resource_path)


class WereSyncWindow(Gtk.Window):
    def __init__(self, title="WereSync"):
        super().__init__(title=title)
        # Find all the bootloader plugins available
        manager = plugins.get_manager()
        manager.collectPlugins()
        plugin_store = Gtk.ListStore(int, str, str)
        plugins_added = []
        uuid_index = 0
        for idx, pluginInfo in enumerate(manager.getAllPlugins()):
            manager.activatePluginByName(pluginInfo.name)
            obj = pluginInfo.plugin_object
            if pluginInfo.name not in plugins_added:
                plugin_store.append([idx, obj.prettyName, obj.name])
                plugins_added.append(pluginInfo.name)
                if obj.name == "uuid_copy":
                    uuid_index = idx
            else:
                LOGGER.debug("Not adding {name} at {path} because plugin"
                             "already added".format(
                                 name=pluginInfo.name, path=pluginInfo.path))

        self.set_icon_from_file(get_resource("weresync.svg"))
        self.grid = Gtk.Grid()
        self.add(self.grid)
        self.source_label = Gtk.Label(
            label=_("Source Drive: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        name_store = Gtk.ListStore(int, str)
        for idx, val in enumerate(generate_drive_list()):
            name_store.append([idx, val])
        self.source_combo = Gtk.ComboBox.new_with_model_and_entry(name_store)
        self.source_combo.set_hexpand(True)
        self.source_combo.set_entry_text_column(1)
        self.grid.attach(self.source_label, 1, 1, 1, 1)
        self.grid.attach_next_to(self.source_combo, self.source_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.target_label = Gtk.Label(
            label=_("Target Drive: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.target_combo = Gtk.ComboBox.new_with_model_and_entry(name_store)
        self.target_combo.set_hexpand(True)
        self.target_combo.set_entry_text_column(1)
        self.grid.attach_next_to(self.target_label, self.source_label,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.target_combo, self.target_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        box = Gtk.Box()
        self.grid.attach_next_to(box, self.source_combo,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.lvm_source_label = Gtk.Label(
            label=_("Source VG: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        lvm_list = generate_vg_list()
        lvm_source_store = Gtk.ListStore(int, str)
        for idx, val in enumerate(lvm_list):
            lvm_source_store.append([idx, val])
        self.lvm_source_combo = Gtk.ComboBox.new_with_model_and_entry(
            lvm_source_store)
        self.lvm_source_combo.set_hexpand(True)
        self.lvm_source_combo.set_entry_text_column(1)
        self.lvm_source_combo.set_sensitive(False)
        self.grid.attach_next_to(self.lvm_source_label, box,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.grid.attach_next_to(self.lvm_source_combo, self.lvm_source_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.lvm_target_label = Gtk.Label(
            label=_("Target VG: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        lvm_target_store = Gtk.ListStore(int, str)
        lvm_target_store.append([1, _("Default")])
        for idx, val in enumerate(lvm_list):
            lvm_target_store.append([idx, val])
        self.lvm_target_combo = Gtk.ComboBox.new_with_model_and_entry(
            lvm_target_store)
        self.lvm_target_combo.set_hexpand(True)
        self.lvm_target_combo.set_entry_text_column(1)
        self.lvm_target_combo.set_active(0)
        self.lvm_target_combo.set_sensitive(False)
        self.grid.attach_next_to(self.lvm_target_label, self.lvm_source_label,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.lvm_target_combo, self.lvm_target_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.copy_partitions_button = Gtk.CheckButton(
            label=_("Copy partitions if target partitions are invalid."))
        set_margin(self.copy_partitions_button)
        self.grid.attach_next_to(self.copy_partitions_button,
                                 self.target_label, Gtk.PositionType.BOTTOM, 2,
                                 1)
        self.lvm_button = Gtk.CheckButton(
            label=_("Copy Logical Volume Groups."))
        self.lvm_button.connect("toggled", self.lvm_button_toggled)
        self.grid.attach_next_to(self.lvm_button, self.lvm_target_label,
                                 Gtk.PositionType.BOTTOM, 2, 1)
        set_margin(self.lvm_button)
        self.bootloader_label = Gtk.Label(
            label=_("Bootloader Plugin: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.bootloader_combo = Gtk.ComboBox.new_with_model_and_entry(
            plugin_store)
        self.bootloader_combo.set_entry_text_column(1)
        self.bootloader_combo.set_active(uuid_index)
        self.bootloader_help = create_help_box(
            self,
            _("This is the plugin which will attempt to make your clone"
              " bootable. Select the plugin which corresponds to the "
              "bootloader"
              " you want to install. If you are unsure what to choose, pick"
              " 'UUID Copy'."), _("Bootloader Plugin"))
        self.grid.attach_next_to(self.bootloader_label,
                                 self.copy_partitions_button,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.bootloader_combo, self.bootloader_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.grid.attach_next_to(self.bootloader_help, self.bootloader_combo,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.bootloader_partition_label = Gtk.Label(
            label=_("Root Partition Number: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.grid.attach_next_to(self.bootloader_partition_label,
                                 self.bootloader_label,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.bootloader_partition_entry = NumberEntry()
        self.grid.attach_next_to(self.bootloader_partition_entry,
                                 self.bootloader_partition_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.bootloader_help = create_help_box(
            self,
            _("Enter the partition number of the partition"
              " to install the bootloader on. This is generally the partition "
              "mounted on /\n"
              "So if your root directory is /dev/sda2, enter 2."),
            _("Bootloader Partition"))
        self.grid.attach_next_to(self.bootloader_help,
                                 self.bootloader_partition_entry,
                                 Gtk.PositionType.RIGHT, 1, 1)
        # Start adding advanced options
        self.boot_part_label = Gtk.Label(
            label=_("Boot Partition: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.grid.attach_next_to(self.boot_part_label, self.lvm_button,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.boot_part_entry = NumberEntry()
        self.grid.attach_next_to(self.boot_part_entry, self.boot_part_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.boot_help = create_help_box(
            self, _("The number of the partition mounted on /boot."),
            _("Boot Partition"))
        self.grid.attach_next_to(self.boot_help, self.boot_part_entry,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.expander = Gtk.Expander(label=_("Advanced Options"))
        self.efi_partition_label = Gtk.Label(
            label=_("EFI Partition Number: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.grid.attach_next_to(self.efi_partition_label,
                                 self.boot_part_label, Gtk.PositionType.BOTTOM,
                                 1, 1)
        self.efi_partition_entry = NumberEntry()
        self.efi_partition_entry.set_hexpand(True)
        self.grid.attach_next_to(self.efi_partition_entry,
                                 self.efi_partition_label,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.efi_help = create_help_box(
            self,
            _("Enter the partition number of your EFI partition.\n"
              "So if your efi partition is found on /dev/sda1,"
              " enter 1.\n"
              "If you are not running a UEFI system, leave this blank."),
            _("EFI Partition"))
        self.grid.attach_next_to(self.efi_help, self.efi_partition_entry,
                                 Gtk.PositionType.RIGHT, 1, 1)
        set_margin(self.expander)
        self.expander.set_resize_toplevel(True)
        self.expand_grid = Gtk.Grid()
        self.expander.add(self.expand_grid)
        self.expander.set_hexpand(True)
        self.ignore_errors = Gtk.CheckButton(
            label=_(
                "Ignore errors during copying. If off, common errors often "
                "stop the clone."))
        self.ignore_errors.set_active(True)
        set_margin(self.ignore_errors)
        self.expand_grid.attach(self.ignore_errors, 1, 1, 3, 1)

        self.source_part_mask_label = Gtk.Label(
            label=_("Source Partition Mask: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.expand_grid.attach_next_to(self.source_part_mask_label,
                                        self.ignore_errors,
                                        Gtk.PositionType.BOTTOM, 1, 1)
        self.source_part_mask_entry = Gtk.Entry()
        self.source_part_mask_entry.set_hexpand(True)
        self.source_part_mask_entry.set_text("{0}{1}")
        self.expand_grid.attach_next_to(self.source_part_mask_entry,
                                        self.source_part_mask_label,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.part_mask_help = create_help_box(
            self,
            _("A string that controls the how partitions are found on the  "
              "file system. It should have two placeholders: "
              "{0} for the device name and {1} for the partition number.\n"
              "So if you have /dev/loop0 and partition 1 is /dev/loop0p1, the "
              "part_mask should be '{0}p{1}'"), _("Partition Mask"))
        self.expand_grid.attach_next_to(self.part_mask_help,
                                        self.source_part_mask_entry,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.target_part_mask_label = Gtk.Label(
            label=_("Target Partition Mask: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.expand_grid.attach_next_to(self.target_part_mask_label,
                                        self.source_part_mask_label,
                                        Gtk.PositionType.BOTTOM, 1, 1)
        self.target_part_mask_entry = Gtk.Entry()
        self.target_part_mask_entry.set_text("{0}{1}")
        self.target_part_mask_entry.set_hexpand(True)
        self.expand_grid.attach_next_to(self.target_part_mask_entry,
                                        self.target_part_mask_label,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.excluded_label = Gtk.Label(
            label=_("Excluded Partitions: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.expand_grid.attach_next_to(self.excluded_label,
                                        self.target_part_mask_label,
                                        Gtk.PositionType.BOTTOM, 1, 1)
        self.excluded_entry = NumberEntry(allowed=", ")
        self.excluded_entry.set_hexpand(True)
        self.expand_grid.attach_next_to(self.excluded_entry,
                                        self.excluded_label,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.excluded_help = create_help_box(
            self,
            _("A comma separated list of partition numbers that should not be "
              "copied or searched.\n"
              "If partitions partitions are copied, they will still be copied."
              ), _("Excluded Partitions"))
        self.expand_grid.attach_next_to(self.excluded_help,
                                        self.excluded_entry,
                                        Gtk.PositionType.RIGHT, 1, 1)

        self.rsync_label = Gtk.Label(
            label=_("Rsync Arguments: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.expand_grid.attach_next_to(self.rsync_label, self.part_mask_help,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.rsync_entry = Gtk.Entry(text=device.DEFAULT_RSYNC_ARGS)
        self.expand_grid.attach_next_to(self.rsync_entry, self.rsync_label,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.rsync_help = create_help_box(
            self,
            _("Enter the arguments to pass the rsync program. For more "
              "information see <a href=\"https://download.samba.org/pub/rsync/rsync.html# Options%20Summary\">the rsync website</a>."  # noqa
              ),  # noqa
            _("Rsync Arguments"))
        self.expand_grid.attach_next_to(self.rsync_help, self.rsync_entry,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.source_mount_label = Gtk.Label(
            label=_("Source Drive Mount Point: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.expand_grid.attach_next_to(self.source_mount_label,
                                        self.rsync_label,
                                        Gtk.PositionType.BOTTOM, 1, 1)
        self.source_mount_entry = Gtk.FileChooserButton(
            title=_("Source Drive Mount Folder"),
            action=Gtk.FileChooserAction.SELECT_FOLDER,
        )
        self.expand_grid.attach_next_to(self.source_mount_entry,
                                        self.source_mount_label,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.mount_help = create_help_box(
            self,
            _("These are the folders that the drives to be copied will be "
              "mounted in. If unset, WereSync will generate random folders in "
              "the /tmp directory. Generally this can be unset."),
            _("Drive Mount Point."))
        self.expand_grid.attach_next_to(self.mount_help,
                                        self.source_mount_entry,
                                        Gtk.PositionType.RIGHT, 1, 1)
        self.target_mount_label = Gtk.Label(
            label=_("Target Drive Mount Point: "))
        self.expand_grid.attach_next_to(self.target_mount_label,
                                        self.source_mount_label,
                                        Gtk.PositionType.BOTTOM, 1, 1)
        self.target_mount_entry = Gtk.FileChooserButton(
            title=_("Target Drive Mount Folder"),
            action=Gtk.FileChooserAction.SELECT_FOLDER)
        self.expand_grid.attach_next_to(self.target_mount_entry,
                                        self.target_mount_label,
                                        Gtk.PositionType.RIGHT, 1, 1)

        # End advanced options
        self.grid.attach_next_to(self.expander,
                                 self.bootloader_partition_label,
                                 Gtk.PositionType.BOTTOM, 6, 1)
        self.start = Gtk.Button(label=_("Start Clone"))
        set_margin(self.start)
        self.start.set_hexpand(False)
        self.grid.attach(self.start, 6, 10, 1, 1)
        self.start.connect("clicked", self.start_pressed)

    def lvm_button_toggled(self, button):
        if self.lvm_button.get_active():
            self.lvm_source_combo.set_sensitive(True)
            self.lvm_target_combo.set_sensitive(True)
        else:
            self.lvm_source_combo.set_sensitive(False)
            self.lvm_target_combo.set_sensitive(False)

    def set_expander(self, val):
        self.expander.set_expanded(val)

    def get_selected_combo(self, combo):
        combo_iter = combo.get_active_iter()
        if combo_iter is not None:
            model = combo.get_model()
            row_id, val = model[combo_iter][:2]
            return val
        else:
            entry = combo.get_child()
            return entry.get_text()

    def start_pressed(self, *args):
        self.source = self.get_selected_combo(self.source_combo)
        self.target = self.get_selected_combo(self.target_combo)
        is_lvm = self.lvm_button.get_active()
        if is_lvm:
            self.lvm_source = self.get_selected_combo(self.lvm_source_combo)
            lvm_target = self.get_selected_combo(self.lvm_target_combo)
            if lvm_target == _("Default"):
                lvm_target = ""
        else:
            self.lvm_source = ""
            lvm_target = ""
        confirm_dialog = Gtk.MessageDialog(self, 0, Gtk.MessageType.QUESTION,
                                           Gtk.ButtonsType.OK_CANCEL, "")
        confirm_dialog.format_secondary_text(
            ("This action will DELETE "
             "everything on {drive} " + (" and {lvm}" if is_lvm else "") +
             ", and make it the "
             "same as the source drives. Is "
             "this what you want to do?").format(
                 drive=self.target, lvm=(lvm_target if is_lvm else "")))
        response = confirm_dialog.run()
        confirm_dialog.destroy()
        if response == Gtk.ResponseType.CANCEL:
            return
        # The user didn't cancel so we can continue running
        copy_if_invalid = self.copy_partitions_button.get_active()
        efi_part = int(self.efi_partition_entry.get_text(
        )) if self.efi_partition_entry.get_text().strip() != "" else -1
        bootloader_part = int(self.bootloader_partition_entry.get_text(
        )) if self.bootloader_partition_entry.get_text() != "" else -1
        ignore_errors = self.ignore_errors.get_active()
        self.source_part_mask = self.source_part_mask_entry.get_text()
        self.target_part_mask = self.target_part_mask_entry.get_text()
        exclude_text = self.excluded_entry.get_text().strip()
        if exclude_text == "":
            excluded_parts = []
        else:
            exclude_text.replace(" ", "")
            excluded_parts = [int(x) for x in exclude_text.split(",")]
        boot_part = int(self.boot_part_entry.get_text()
                        ) if self.boot_part_entry.get_text() != "" else -1
        rsync_args = self.rsync_entry.get_text()
        source_mount = self.source_mount_entry.get_filename()
        target_mount = self.target_mount_entry.get_filename()
        mount_points = (source_mount if source_mount is not None else "",
                        target_mount if target_mount is not None else "")
        boot_iter = self.bootloader_combo.get_active_iter()
        model = self.bootloader_combo.get_model()
        plugin_name = model[boot_iter][2]
        try:
            self._generate_progress_grid()
            self.remove(self.grid)
            self.add(self.progress_grid)

            dbus_client.subscribe_to_signals(
                lambda x: GLib.idle_add(self.part_callback, x),
                lambda num, prog: GLib.idle_add(self.copy_callback, num, prog),
                lambda done: GLib.idle_add(self.boot_callback, done))

            def copy(callback, error):
                try:
                    result = dbus_client.copy_drive(
                        self.source, self.target, copy_if_invalid,
                        self.source_part_mask, self.target_part_mask,
                        excluded_parts, ignore_errors, bootloader_part,
                        boot_part, efi_part, mount_points, rsync_args,
                        self.lvm_source, lvm_target, plugin_name)
                    callback(result)
                except Exception as ex:
                    LOGGER.debug(
                        "Full exception info:\n", exc_info=sys.exc_info())
                    error(ex)

            copy_thread = threading.Thread(
                target=copy,
                args=[
                    lambda result: GLib.idle_add(self._copy_finished, result),
                    lambda ex: GLib.idle_add(self._show_error, ex)
                ])
            copy_thread.start()
            self.show_all()
        except Exception as ex:
            LOGGER.debug("Full exception info:\n", exc_info=sys.exc_info())
            self._show_error(ex)
            return

    def _show_error(self, ex):
        """Displays an error in a message dialog."""
        dialog = Gtk.MessageDialog(self, 0, Gtk.MessageType.ERROR,
                                   Gtk.ButtonsType.OK,
                                   _("Error starting clone."))
        dialog.format_secondary_text(str(ex))
        dialog.run()
        dialog.destroy()

        # Sets back to original screen to allow regenerating any misplace
        # parameters.
        self.remove(self.progress_grid)
        self.add(self.grid)

    def _copy_finished(self, result):
        """A callback function to be run when the the drive finishes
        copying."""
        text = _("Clone finished!")
        if result is not True:
            text += _("\nNon fatal error occurred: ") + str(result)

        dialog = Gtk.MessageDialog(self, 0, Gtk.MessageType.INFO,
                                   Gtk.ButtonsType.OK, text)
        dialog.run()
        dialog.destroy()
        self.remove(self.progress_grid)
        self.add(self.grid)

    def _generate_progress_grid(self):
        """Generates the grid for the screen showing progress. Sets
        `self.progress_grid` as the grid.`"""

        self.progress_grid = Gtk.Grid()
        part_label = Gtk.Label(
            label=_("Checking partitions and copying: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.progress_grid.attach(part_label, 1, 1, 1, 1)
        self.part_progress = Gtk.ProgressBar()
        set_margin(self.part_progress)
        self.progress_grid.attach_next_to(self.part_progress, part_label,
                                          Gtk.PositionType.RIGHT, 1, 1)
        self.copy_progresses = {}

        def create_partitions(device, device_part_mask, start_label,
                              lvm=False):
            previous_label = start_label
            partitions = dbus_client.drive_copier.GetPartitions(
                device, device_part_mask, lvm)
            for val in partitions:
                copy_label = Gtk.Label(
                    label=_("Copying partition {0}: ").format(val),
                    halign=Gtk.Align.START,
                    xpad=DEFAULT_HORIZONTAL_PADDING,
                    ypad=DEFAULT_VERTICAL_PADDING)
                copy_progress = Gtk.ProgressBar()
                set_margin(copy_progress)
                self.progress_grid.attach_next_to(
                    copy_label, previous_label, Gtk.PositionType.BOTTOM, 1, 1)
                self.progress_grid.attach_next_to(copy_progress, copy_label,
                                                  Gtk.PositionType.RIGHT, 1, 1)
                self.copy_progresses[val] = copy_progress
                previous_label = copy_label

            return previous_label

        final_label = create_partitions(self.source, self.source_part_mask,
                                        part_label)
        if self.lvm_button.get_active():
            final_label = create_partitions(
                self.lvm_source, "", final_label, lvm=True)

        boot_label = Gtk.Label(
            label=_("Making bootable: "),
            halign=Gtk.Align.START,
            xpad=DEFAULT_HORIZONTAL_PADDING,
            ypad=DEFAULT_VERTICAL_PADDING)
        self.progress_grid.attach_next_to(boot_label, final_label,
                                          Gtk.PositionType.BOTTOM, 1, 1)
        self.boot_progress = Gtk.ProgressBar()
        set_margin(self.boot_progress)
        self.progress_grid.attach_next_to(self.boot_progress, boot_label,
                                          Gtk.PositionType.RIGHT, 1, 1)
        self.cancel_btn = Gtk.Button(label="Cancel")
        set_margin(self.cancel_btn)
        self.progress_grid.attach_next_to(self.cancel_btn, self.boot_progress,
                                          Gtk.PositionType.BOTTOM, 1, 1)

    def part_callback(self, progress):
        LOGGER.debug("part callback. Value: {0}".format(progress))
        self.part_progress.set_fraction(progress)

    def copy_callback(self, part, progress):
        part = str(int(part))
        if progress < 0:
            LOGGER.debug(
                "Error occurred copying partition {0}. Marking complete.".
                format(part))
            self.copy_progresses[part].set_fraction(1.0)
        elif progress is True and isinstance(progress, bool):
            self.copy_progresses[part].pulse()
        elif (progress >= self.copy_progresses[part].get_fraction()):
            self.copy_progresses[part].set_fraction(progress)

    def boot_callback(self, done):
        if not done:
            self.boot_progress.pulse()
        else:
            self.boot_progress.set_fraction(1.0)


def start_gui():

    utils.enable_localization()

    try:
        utils.check_python_version()
    except InvalidVersionError as ex:
        print(ex)
        # This might not work, if the user doesn't have a setup that support
        # PyGObject, but it's worth a shot.
        dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.ERROR,
                                   Gtk.ButtonsType.OK,
                                   _("Error starting WereSync."))
        dialog.format_secondary_text(str(ex))
        dialog.run()
        dialog.destroy()
        sys.exit(1)

    utils.start_logging_handler(utils.DEFAULT_USER_LOG_LOCATION)

    if not connected:
        dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.ERROR,
                                   Gtk.ButtonsType.OK,
                                   _("Error starting WereSync."))
        dialog.format_secondary_text(_("Weresync service not connected."))
        dialog.run()
        dialog.destroy()
        sys.exit(1)

    LOGGER.info(_("Starting gui."))
    GObject.threads_init()
    win = WereSyncWindow()
    win.connect("delete-event", Gtk.main_quit)
    # This is set to expanded so it will be centered as if advanced options
    # were opened
    win.set_expander(True)
    win.set_position(Gtk.WindowPosition.CENTER)
    win.show_all()
    # Then advanced options are closed so as not to be distracting
    win.set_expander(False)
    win.show_all()
    Gtk.main()
